"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import hashlib
import re
from collections.abc import AsyncGenerator, Mapping, Sequence
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any, Final, Generic, NamedTuple, cast
from zlib import crc32

from typing_extensions import Self, TypeVar

from . import utils
from ._const import (
    JSON_LOADS,
    STATE,
    STEAM_BADGES,
    UNIX_EPOCH,
    URL,
    WRITE_U32,
    ReadOnly,
    impl_eq_via_id,
)
from .achievement import AppAchievement, AppStats
from .badge import AppBadge
from .enums import *
from .id import _ID64_TO_ID32, id64_from_url
from .models import CDNAsset, DescriptionMixin, _IOMixin
from .protobufs import client_server, player
from .tag import Category, Tag
from .types.id import *
from .utils import DateTime

if TYPE_CHECKING:
    from .abc import PartialUser
    from .clan import Clan
    from .friend import Friend
    from .leaderboard import Leaderboard
    from .manifest import AppInfo, Depot, HeadlessDepot, Manifest
    from .package import FetchedAppPackage, License, PartialPackage
    from .protobufs import econ
    from .protobufs.encrypted_app_ticket import EncryptedAppTicket as EncryptedAppTicketProto
    from .published_file import PublishedFile
    from .review import Review
    from .state import ConnectionState
    from .store import AppStoreItem
    from .types import app
    from .types.user import Author

__all__ = (
    "App",
    "PartialApp",
    "TF2",
    "LFD2",
    "DOTA2",
    "CSGO",
    "STEAM",
    "CUSTOM_APP",
    "OwnershipTicket",
    "AuthenticationTicketVerificationResult",
    "AuthenticationTicket",
    "EncryptedTicket",
    "FriendThoughts",
    "AppShopItem",
    "AppShopItemTag",
    "AppShopItems",
    "CommunityItem",
    "RewardItem",
    "DLC",
    "UserApp",
    "UserRecentlyPlayedApp",
    "WishlistApp",
    "FetchedApp",
    "UserInventoryInfoApp",
    "UserInventoryInfoContext",
)

T = TypeVar("T")
APP_ID_MAX: Final = AppID((1 << 32) - 1)
NameT = TypeVar("NameT", bound=str | None, default=str | None, covariant=True)


@impl_eq_via_id
class App(Generic[NameT]):
    """Represents a Steam app."""

    __slots__ = (
        "id",
        "name",
    )

    def __init__(
        self,
        *,
        id: Intable,
        name: NameT = None,
    ):
        try:
            id = int(id)
        except (ValueError, TypeError):
            raise ValueError("id expected to support int()") from None

        if id < 0:
            raise ValueError("id cannot be negative")

        self.id: AppID = AppID(id)
        """The app's app ID."""
        self.name = name
        """The app's name."""

    def __str__(self) -> str:
        return self.name or f"App: {self.id}"

    def __repr__(self) -> str:
        attrs = ("name", "id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"{self.__class__.__name__}({', '.join(resolved)})"

    def to_proto(self) -> client_server.CMsgClientGamesPlayedGamePlayed:
        if self.is_valid():
            return client_server.CMsgClientGamesPlayedGamePlayed(game_id=self.id)

        if self.name is None:
            raise TypeError("un-serializable app with no name")
        return client_server.CMsgClientGamesPlayedGamePlayed(game_id=self.id, game_extra_info=self.name)

    def is_valid(self) -> bool:
        """Whether the app could be a Steam app."""
        return self.id <= APP_ID_MAX

    @property
    def url(self) -> str:
        """The app's URL on https://steamcommunity.com."""
        return f"{URL.COMMUNITY}/app/{self.id}"


def CUSTOM_APP(
    name: str,
) -> App[str]:
    """Create a custom app instance for :meth:`~steam.Client.change_presence`.
    The :attr:`App.id` will be set to ``15190414816125648896``.

    Example:

    .. code:: python

        await client.change_presence(app=steam.CUSTOM_APP("my cool game"))

    Parameters
    ----------
    name
        The name of the app to set your playing status to.
    """
    # if actually optimising make this return a different class cause it's a u64 cause haha steam
    return App(name=name, id=15190414816125648896)


class BaseOwnershipTicket:
    __slots__ = (
        "_state",
        "_ticket",
        "_start",
        "_end",
        "version",
        "user",
        "app",
        "external_ip",
        "internal_ip",
        "flags",
        "created_at",
        "_expires",
        "licenses",
        "dlc",
        "signature",
    )

    def __init__(self, state: ConnectionState, ticket: utils.StructIO) -> None:
        from .package import PartialPackage

        self._state = state
        self._ticket = ticket
        self._start = ticket.position
        self._end = ticket.read_u32()  # including itself, for some reason
        if self._start + self._end != len(ticket) and self._start + self._end + 128 != len(ticket):
            return None

        self.version = ticket.read_u32()
        """The version of the ticket."""
        self.user = state.get_partial_user(ticket.read_u64())
        """The user who owns the ticket."""
        self.app = PartialApp(state, id=ticket.read_u32())
        """The app the ticket is for."""
        self.external_ip = IPv4Address(ticket.read_u32())
        """The external IP address of the user."""
        self.internal_ip = IPv4Address(ticket.read_u32())
        """The internal IP address of the user."""
        self.flags = ticket.read_u32()
        """The flags of the ticket."""
        self.created_at = DateTime.from_timestamp(ticket.read_u32())
        """The time the ticket was created."""
        self._expires = DateTime.from_timestamp(ticket.read_u32())
        """The time the ticket expires."""
        self.licenses = [PartialPackage(state, id=ticket.read_u32()) for _ in range(ticket.read_u16())]
        """The licenses the user owns."""
        self.dlc = [
            OwnershipDLC(
                state,
                id=ticket.read_u32(),
                owned_packages=[PartialPackage(state, id=ticket.read_u32()) for _ in range(ticket.read_u16())],
            )
            for _ in range(ticket.read_u16())
        ]
        """The DLC the user owns."""

        ticket.read_u16()  # reserved
        signature = ticket.read(128)
        self.signature = signature if len(signature) == 128 else None
        """The signature of the ticket."""

    def __repr__(self) -> str:
        attrs = (
            "user",
            "app",
            "version",
        )
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class OwnershipTicket(BaseOwnershipTicket):
    """Represents an ownership ticket. This is used to verify ownership of an app."""

    __slots__ = ()

    def __bytes__(self) -> bytes:
        return self._ticket.buffer

    def is_signature_valid(self) -> bool:
        if self.signature is not None:
            return utils.verify_signature(
                self._ticket.getbuffer()[self._start : self._start + self._end],
                self.signature,
            )
        return False

    @property
    def expires(self) -> datetime:
        """The time at which the ticket expires."""
        return self._expires

    def is_expired(self) -> bool:
        """Whether the ticket has expired."""
        return self._expires < DateTime.now()

    def is_valid(self) -> bool:
        return not self.is_expired() and (not self.signature or self.is_signature_valid())


@dataclass(slots=True)
class AuthenticationTicketVerificationResult:
    """Represents the result of an authentication ticket verification.

    .. container:: operations

        .. describe:: bool(x)

            Checks if the result is :attr:`Result.OK`.
    """

    result: Result
    """The result of the verification."""
    user: PartialUser
    """The user who sent the ticket."""
    owner: PartialUser
    """The user who owns the ticket."""
    vac_banned: bool
    """Whether the user is VAC banned."""
    publisher_banned: bool
    """Whether the user is publisher banned."""

    def __bool__(self) -> bool:
        return self.result is Result.OK


class AuthenticationTicket(OwnershipTicket):
    """Represents an authentication ticket. This is used to verify ownership of an app and to connect to the game server."""

    __slots__ = (
        "auth_ticket",
        "gc_token",
        "gc_token_created_at",
        "client_ip",
        "client_connected_at",
        "client_connection_count",
    )

    def __init__(self, state: ConnectionState, ticket: utils.StructIO) -> None:
        self.auth_ticket = bytes(ticket.getbuffer()[ticket.position - 4 : ticket.position - 4 + 52])
        """The authentication ticket for the app. The first 52 bytes of the ticket."""
        # this is the part that's passed back to Steam for validation

        self.gc_token = ticket.read_u64()
        """The Game Connect token for the app."""
        ticket.position += 8
        self.gc_token_created_at = DateTime.from_timestamp(ticket.read_u32())
        """When the Game Connect token was created."""

        if ticket.read_u32() != 24:
            raise ValueError("Invalid session header")

        ticket.position += 8  # unknown 1 and unknown 2
        self.client_ip = IPv4Address(ticket.read_u32())
        """The IP address of the client."""
        ticket.position += 4  # filler
        self.client_connected_at = timedelta(milliseconds=ticket.read_u32())
        """The time the client has been connected to Steam"""
        self.client_connection_count = ticket.read_u32()
        """How many servers the client has connected to"""

        if ticket.read_u32() + ticket.position != len(ticket):
            raise ValueError("Invalid ownership section")
        super().__init__(state, ticket)

    def __bytes__(self, _header: bytearray = bytearray(WRITE_U32(20))) -> bytes:
        return bytes(_header + self._ticket.getbuffer())

    async def verify(self, *, publisher_api_key: str | None = None) -> AuthenticationTicketVerificationResult:
        """Verify the ticket with the web API.

        Parameters
        ----------
        publisher_api_key
            The publisher API key to use for verification. If not provided, will use the standard (rate limited) API.
        """
        data = await self._state.http.verify_app_ticket(self.app.id, self._ticket.buffer.hex(), publisher_api_key)
        return AuthenticationTicketVerificationResult(
            Result[data["result"]],
            *await self._state._maybe_users((ID64(int(data["steamid"])), ID64(int(data["ownersteamid"])))),
            vac_banned=data["vacbanned"],
            publisher_banned=data["publisherbanned"],
        )

    async def activate(self) -> None:
        """Activate the ticket."""
        await self._state.activate_auth_session_tickets(self)

    async def deactivate(self) -> None:
        """Deactivate the ticket. Ends our sessions with other users meaning we can't get events from them anymore."""
        await self._state.deactivate_auth_session_tickets(self)

    # def is_valid(self) -> bool:  # TODO is it worth having an opinion on this?
    #     return


class EncryptedTicket(BaseOwnershipTicket):
    """Represents an encrypted ticket."""

    def __init__(self, state: ConnectionState, ticket: EncryptedAppTicketProto, key: bytes) -> None:
        decrypted = utils.StructIO(utils.symmetric_decrypt(ticket.encrypted_ticket, key))
        if crc32(decrypted.getbuffer()) != ticket.crc_encryptedticket:
            raise ValueError("Invalid CRC")
        self.user_data = decrypted.read(ticket.cb_encrypteduserdata)
        """The user's given data for the ticket."""

        (length,) = decrypted.read_struct(">I")

        super().__init__(state, utils.StructIO(decrypted.read(length)))
        remaining = decrypted.read()
        if len(remaining) >= 8 + 20:
            to_hash = decrypted.buffer[: ticket.cb_encrypteduserdata + length]
            salt = remaining[:8]
            hash = remaining[8:28]

            if hashlib.sha1(to_hash + salt).digest() != hash:
                raise ValueError("Invalid hash")


async def parse_app_ticket(state: ConnectionState, ticket: utils.StructIO) -> OwnershipTicket | AuthenticationTicket:
    if ticket.read_u32() == 20:
        app_ticket = AuthenticationTicket(state, ticket)
    else:
        ticket.position -= 4
        app_ticket = OwnershipTicket(state, ticket)

    if not app_ticket.is_valid():
        raise ValueError("Invalid ticket")
    app_ticket.user = await state._maybe_user(app_ticket.user.id64)
    return app_ticket


class FriendThoughts(NamedTuple):
    recommended: list[Friend]
    not_recommended: list[Friend]


@dataclass(slots=True)
class AppShopItemTag:
    name: str
    display_name: str
    id: int


class AppShopItem(DescriptionMixin):
    """Represents an item that can be purchased from the shop."""

    __slots__ = (
        *DescriptionMixin.SLOTS,
        "def_index",
        "class_",
        "prices",
        "original_prices",
        "updated_at",
        "store_tags",
    )

    def __init__(
        self,
        state: ConnectionState,
        data: app.AssetPricesAsset,
        description: econ.ItemDescription,
        tags: Sequence[AppShopItemTag],
    ):
        self._state = state
        self.class_id = ClassID(description.classid)
        self.def_index = int(data["name"])
        """The def index of the item in the app's schema"""
        self.class_ = data["class"]
        """Extra info about the item."""
        self.prices = cast(
            Mapping[Currency, int],
            {Currency.try_name(name): price for name, price in data["prices"].items()},
        )
        """The prices of the asset in the store."""
        self.original_prices = cast(
            Mapping[Currency, int] | None,
            (
                {
                    Currency.try_name(name): price
                    for name, price in data["original_prices"].items()
                    if data["prices"][name] < price
                }
                if "original_prices" in data
                else None
            ),
        )
        """The original prices of any items if the price in ``prices`` is reduced."""
        try:
            self.updated_at = DateTime.strptime(
                data["date"], "%Y/%m/%d"
            )  # yes this could be just a date but maybe volvo will be nice one day
            """The time the price was last updated"""
        except ValueError:
            self.updated_at = None  # they like to sprinkle in a bit of 1960/00/00 for the funsies
        self.store_tags = [tag for tag in tags for tag_id in data.get("tag_ids", ()) if tag.id == tag_id]
        """The tags associated with the item."""

        super().__init__(state, description)

    def __repr__(self) -> str:
        attrs = ("name", "class_id", "def_index", "store_tags", "app")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


@dataclass(slots=True)
class AppShopItems(Sequence[AppShopItem]):
    """Represents the items that can be purchased from the shop.

    .. container:: operations

        .. describe:: x == y

            Checks if two app shop items are equal.

        .. describe:: len(x)

            Returns the number of items in the shop.

        .. describe:: x[i]

            Returns the item at index ``i``.
    """

    items: Sequence[AppShopItem]
    """The items that can be purchased from the shop"""
    tags: Sequence[AppShopItemTag]
    """All the possible tags for ``items`` to have."""

    if not TYPE_CHECKING:

        def __len__(self):
            return len(self.items)

        def __getitem__(self, idx):
            return self.items[idx]


AppT = TypeVar("AppT", bound="PartialApp", covariant=True)


@dataclass(slots=True)
class CommunityItem(Generic[AppT]):
    """Represents a community item.

    .. container:: operations

        .. describe:: x == y

            Checks if two community items are equal.

        .. describe:: hash(x)

            Returns the community item's hash.
    """

    _state: ConnectionState = field(repr=False, compare=False)
    type: int
    """The type of the community item."""
    app: AppT
    """The app the community item is from."""
    name: str
    """The name of the community item."""
    title: str
    """The title of the community item."""
    description: str
    """The description of the community item."""
    image: CDNAsset | None
    """The image of the community item. Attempts to use the larger image if possible."""
    movie: CDNAsset | None
    """The movie of the community item. Uses the ``.mp4`` version."""
    data: Mapping[str, Any] | None
    """Data associated with the community item."""
    series: int
    """The series of the community item."""
    class_: CommunityItemClass
    """The class of the community item."""
    editor: Author | None
    """The last editor of the community item."""
    active: bool
    """Whether the community item is active."""
    image_composed: CDNAsset | None
    """The composed image of the community item."""
    image_composed_foil: CDNAsset | None
    """The composed foil image of the community item."""
    deleted: bool
    """Whether the community item is deleted."""
    edited_at: datetime | None
    """When the community item was last edited."""
    # broadcast_channel_id: int = betterproto.uint64_field(17)

    def __hash__(self) -> int:
        hashable_slots = set(self.__slots__) - {"data"}
        return hash(
            (*tuple(getattr(self, attr) for attr in hashable_slots), tuple(self.data.items()) if self.data else None)
        )

    @property
    def badges(self) -> Sequence[AppBadge[AppT]]:
        """The badges for the item."""
        if self.class_ is not CommunityItemClass.Badge:
            return []
        if not self.data:
            return []

        badges: list[AppBadge[AppT]] = []
        previous_name = ""
        for name, image in zip(self.data["level_names"].values(), self.data["level_images"].values()):
            if not name:
                name = previous_name
            badge = AppBadge(
                self._state,
                1,
                name,
                self.app,
                CDNAsset(self._state, f"{URL.CDN}/steamcommunity/public/images/items/{self.app.id}/{image}"),
            )
            previous_name = name
            badges.append(badge)
        return badges


@dataclass(slots=True, unsafe_hash=True)
class RewardItem(Generic[AppT]):
    """Represents a reward item in the Steam Points Shop."""

    type: int
    """The type of the reward item."""
    app: AppT
    """The app the reward item is from."""
    name: str
    """The name of the reward item."""
    title: str
    """The title of the reward item."""
    description: str
    """The description of the reward item."""
    display_description: str
    """The displayed description of the reward item."""
    image: CDNAsset | None
    """The image of the reward item. Attempts to use the larger image if possible."""
    movie: CDNAsset | None
    """The movie of the reward item. Uses the ``.mp4`` version."""
    animated: bool
    """Whether the reward item is animated."""
    badges: Sequence[AppBadge[AppT]]
    """The badges for the item."""
    def_index: int
    """The def index of the reward item."""
    quantity: int
    """The quantity of the reward item."""
    class_: CommunityItemClass
    """The class of the reward item."""
    item_type: int  # 6 is bundle, 1 is normal?
    """The type of the reward item."""
    point_cost: int
    """The cost of the reward item in Steam Points."""
    created_at: datetime
    """When the reward item was created."""
    updated_at: datetime
    """When the reward item was last updated."""
    available_at: datetime
    """When the reward item was made available."""
    availability_ends: datetime
    """When the reward item is no longer available."""
    active: bool
    """Whether the reward item is active."""
    profile_theme_id: str
    """The profile theme ID of the reward item."""
    usable_duration: timedelta
    """How long the reward item is usable for."""
    bundle_discount: int
    """The discount of the reward item if it is a bundle."""
    bundles: Sequence[Self] = field(init=False)
    """The bundles the reward contains."""


class PartialApp(App[NameT]):
    """Apps that have state."""

    __slots__ = ("_state",)

    def __init__(self, state: ConnectionState, *, id: Intable, name: NameT = None):
        super().__init__(id=id, name=name)
        self._state = state

    def __repr__(self) -> str:
        attrs = ("name", "id")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    async def clan(self) -> Clan:
        """Fetch this app's clan.

        This can be useful to get an App's updates.

        .. code:: python

            clan = await app.clan()
            async for update in clan.announcements():
                if update.type in (steam.EventType.MajorUpdate, steam.EventType.SmallUpdate):
                    ...  # do something with the update

        Raises
        ------
        ValueError
            This app has no associated clan.
        """

        id64 = await id64_from_url(self.url, self._state.http._session)
        if id64 is None:
            raise ValueError("App has no associated clan")
        return await self._state.fetch_clan(id64)

    async def player_count(self) -> int:
        """The apps current player count."""
        return await self._state.fetch_app_player_count(self.id)

    async def stats(self, *, language: Language | None = None) -> AppStats:
        """The stats for this app.

        Parameters
        ----------
        language
            The language to fetch the stats in. If ``None``, the current language will be used.

        See Also
        --------
        :meth:`achievements` if you want a faster way to fetch just achievements.
        """
        data = await self._state.http.get_app_stats(self.id, language)
        return AppStats(self._state, self, data)

    async def achievements(self, *, language: Language | None = None) -> list[AppAchievement]:
        """The achievements for this app.

        Parameters
        ----------
        language
            The language to fetch the achievements in. If ``None``, the current language will be used.
        """
        achievements = await self._state.fetch_app_achievements(self.id, language)
        return [
            AppAchievement(
                achievement.internal_name,
                self,
                achievement.localized_name,
                achievement.localized_desc,
                CDNAsset(
                    self._state,
                    f"{URL.CDN}/steamcommunity/public/images/apps/{self.id}/{achievement.icon}",
                ),
                CDNAsset(self._state, f"{URL.CDN}/steamcommunity/public/images/apps/{self.id}/{achievement.icon_gray}"),
                achievement.hidden,
                float(achievement.player_percent_unlocked),
            )
            for achievement in achievements
        ]

    async def leaderboards(self, *, language: Language | None = None) -> list[Leaderboard[Self, str]]:
        """The leaderboards for this app.

        Parameters
        ----------
        language
            The language to fetch the leaderboards in. If ``None``, the current language will be used.
        """
        from .leaderboard import Leaderboard

        leaderboards = await self._state.http.get_app_leaderboards(self.id, language)
        return [
            Leaderboard(
                self._state,
                id=LeaderboardID(leaderboard["id"]),
                app=self,
                name=leaderboard["name"],
                display_name=leaderboard["display_name"],
                entry_count=leaderboard["entry_count"],
                sort_method=LeaderboardSortMethod.try_value(leaderboard["sort_method"]),
                display_type=LeaderboardDisplayType.try_value(leaderboard["display_type"]),
            )
            for leaderboard in leaderboards
        ]

    async def fetch_leaderboard(self, name: str) -> Leaderboard[Self, None]:
        """Fetch a leaderboard by name.

        Parameters
        ----------
        name
            The name of the leaderboard to fetch.

            Note
            ----
            This is not the name of the leaderboard shown in the app's stat page, you can find the name of the leaderboard
            from :meth:`leaderboards`.
        """
        from .leaderboard import Leaderboard

        leaderboard = await self._state.fetch_or_create_app_leaderboard(self.id, name)
        return Leaderboard(
            self._state,
            id=LeaderboardID(int(leaderboard.leaderboard_id)),
            app=self,
            name=leaderboard.leaderboard_name,
            entry_count=leaderboard.leaderboard_entry_count,
            sort_method=LeaderboardSortMethod.try_value(leaderboard.leaderboard_sort_method),
            display_type=LeaderboardDisplayType.try_value(leaderboard.leaderboard_display_type),
        )

    async def create_leaderboard(self, name: str) -> Leaderboard[Self, None]:
        """Create a leaderboard with a given name.

        Parameters
        ----------
        name
            The name of the leaderboard to create.
        """
        from .leaderboard import Leaderboard

        leaderboard = await self._state.fetch_or_create_app_leaderboard(self.id, name, create=True)
        return Leaderboard(
            self._state,
            id=LeaderboardID(int(leaderboard.leaderboard_id)),
            app=self,
            name=leaderboard.leaderboard_name,
            entry_count=leaderboard.leaderboard_entry_count,
            sort_method=LeaderboardSortMethod.try_value(leaderboard.leaderboard_sort_method),
            display_type=LeaderboardDisplayType.try_value(leaderboard.leaderboard_display_type),
        )

    async def friends_who_own(self) -> list[Friend]:
        """Fetch the users in your friend list who own this app."""
        id64s = await self._state.fetch_friends_who_own(self.id)
        return [self._state.get_friend(_ID64_TO_ID32(id64)) for id64 in id64s]

    async def review(
        self,
        content: str,
        *,
        recommend: bool,
        public: bool = True,
        commentable: bool = True,
        received_compensation: bool = False,
        language: Language | None = None,
    ) -> Review:
        """Review an app.

        Parameters
        ----------
        content
            The content of the review.
        recommend
            Whether you recommended the app.
        public
            Whether the review should be public.
        commentable
            Whether the review should allow comments.
        received_compensation
            Whether you received compensation for this review.
        language
            The language the review is in.
        """
        language = language or self._state.http.language
        await self._state.http.post_review(
            self.id, content, recommend, public, commentable, received_compensation, language.api_name
        )
        return await self._state.user.fetch_review(self)  # TODO this sucks can we actually get the id ourselves?

    async def reviews(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[Review, None]:
        """An :term:`asynchronous iterator` for accessing a :class:`steam.App`'s
        :class:`steam.Review`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for review in app.reviews(limit=10):
                print("Reviewer:", review.author)
                print("Said:", review.content)

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of reviews to search through. Default is ``100``. Setting this to ``None`` will fetch all
            the app's reviews, but this will be a very slow operation.
        before
            A time to search for reviews before.
        after
            A time to search for reviews after.

        Yields
        ------
        :class:`~steam.Review`
        """
        from .review import Review, ReviewApp

        after = after or UNIX_EPOCH
        before = before or DateTime.now()
        cursor = "*"
        app = None

        yielded = 0

        while True:
            data = await self._state.http.get_reviews(self.id, "all", "all", "all", cursor)
            if cursor == "*":
                app = ReviewApp(self._state, self.id, data["query_summary"]["review_score"])
            assert app
            cursor = data["cursor"]
            reviews = data["reviews"]

            for review, user in zip(
                reviews,
                await self._state.fetch_users(ID64(int(review["author"]["steamid"])) for review in reviews),
            ):
                review = Review._from_data(self._state, review, app, user)
                if not after < review.created_at < before:
                    return
                if limit is not None and yielded >= limit:
                    return

                yield review
                yielded += 1

    async def friend_thoughts(self) -> FriendThoughts:
        """Fetch the client user's friends who recommended and didn't recommend this app in a review.

        .. source:: FriendThoughts
        """
        proto = await self._state.fetch_friend_thoughts(self.id)
        return FriendThoughts(
            [self._state.get_friend(id) for id in cast("list[ID32]", proto.accountids_recommended)],
            [self._state.get_friend(id) for id in cast("list[ID32]", proto.accountids_not_recommended)],
        )

    async def fetch(self, *, language: Language | None = None) -> FetchedApp:
        """Fetch this app.

        Shorthand for:

        .. code:: python

            app = await client.fetch_app(app_id, language=language)
        """
        return await self._state.fetch_app(self.id, language=language)

    async def store_item(self, *, language: Language | None = None) -> AppStoreItem:
        """Fetch the store item for this app.

        Shorthand for:

        .. code:: python

            (item,) = await client.fetch_store_item(apps=[app], language=language)
        """
        from .store import AppStoreItem

        language = language or self._state.http.language
        (item,) = await self._state.fetch_store_info(app_ids=(self.id,), language=language)
        return AppStoreItem(self._state, item, language)

    async def info(self) -> AppInfo:
        """Fetches this app's product info.

        Shorthand for:

        .. code:: python

            (info,) = await client.fetch_product_info(apps=[app])
        """
        (info,), _ = await self._state.fetch_product_info((self.id,))
        return info

    async def depots(self) -> Sequence[Depot | HeadlessDepot]:
        """Fetch the depots for this app."""
        info = await self.info()
        return await info.depots()

    async def fetch_manifest(
        self, *, id: int, depot_id: int, branch: str = "public", password_hash: str = ""
    ) -> Manifest:
        """Fetch a CDN manifest for an app.

        Parameters
        ----------
        id
            The ID of the manifest to fetch.
        depot_id
            The ID of the manifest's associated depot.
        branch
            The name of the branch the manifest is from.
        password_hash
            The hashed password for the manifest.
        """
        return await self._state.fetch_manifest(
            self.id, ManifestID(id), DepotID(depot_id), name=None, branch=branch, password_hash=password_hash
        )

    async def manifests(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
        branch: str = "public",
        password: str | None = None,
        password_hash: str = "",
    ) -> AsyncGenerator[Manifest, None]:
        """An :term:`asynchronous iterator` for accessing a this app's :class:`steam.Manifest`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for manifest in app.manifests(limit=10):
                print("Manifest:", manifest.name)
                print(f"Contains {len(manifest.paths)} manifests")

        All parameters are optional.

        Parameters
        ----------
        limit
            The maximum number of :class:`.Manifests` to return.
        before
            The time to get manifests before.
        after
            The time to get manifests after.
        branch
            The name of the branch to fetch manifests from.
        password
            The password for the branch, if any.
        password_hash
            The hashed password for a manifest.

        Yields
        ------
        :class:`Manifest`
        """
        after = after or UNIX_EPOCH
        before = before or DateTime.now()

        manifest_coros = await self._state.fetch_manifests(self.id, branch, password, limit, password_hash)
        for chunk in utils.as_chunks(manifest_coros, 100):
            for manifest in await asyncio.gather(*chunk):
                if after < manifest.created_at < before:
                    yield manifest

    async def published_files(
        self,
        *,
        type: PublishedFileQueryFileType = PublishedFileQueryFileType.Items,
        revision: PublishedFileRevision = PublishedFileRevision.Default,
        language: Language | None = None,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[PublishedFile, None]:
        """An :term:`asynchronous iterator` for accessing an app's :class:`steam.PublishedFile`\\s.

        Examples
        --------

        Usage:

        .. code:: python

            async for published_file in app.published_files(limit=10):
                print("Published file:", published_file.name)
                print("Published at:", published_file.created_at)
                print("Published by:", published_file.author)

        All parameters are optional.

        Parameters
        ----------
        type
            The type of published files to fetch.
        revision
            The desired revision of the published files to fetch.
        language
            The language to fetch the published files in. If ``None`` the current language is used.
        limit
            The maximum number of published files to search through. Default is ``100``. Setting this to ``None`` will
            fetch all the app's published files, but this will be a very slow operation.
        before
            A time to search for published files before.
        after
            A time to search for published files after.

        Yields
        ------
        :class:`~steam.PublishedFile`
        """
        from .published_file import PublishedFile

        before = before or DateTime.now()
        after = after or UNIX_EPOCH
        remaining = None
        cursor = "*"
        yielded = 0

        while remaining is None or remaining > 0:
            protos = await self._state.fetch_app_published_files(self.id, type, revision, language, limit, cursor)
            if remaining is None:
                remaining = protos.total
            remaining -= len(protos.publishedfiledetails)
            cursor = protos.next_cursor

            files: list[PublishedFile] = []
            for file in protos.publishedfiledetails:
                file = PublishedFile(self._state, file, self._state.get_partial_user(file.creator))
                if not after < file.created_at < before:
                    remaining = 0
                    break
                files.append(file)

            for file, author in zip(files, await self._state._maybe_users(file.author.id64 for file in files)):
                if limit is not None and yielded >= limit:
                    return
                file.author = author
                yield file
                yielded += 1

    async def dlc(self, *, language: Language | None = None) -> list[DLC]:
        """Fetch the app's DLC.

        Parameters
        ----------
        language
            The language to fetch the DLC in. If ``None``, the current language will be used.
        """
        data = await self._state.http.get_app_dlc(self.id, language)
        return [DLC(self._state, dlc) for dlc in data["dlc"]]

    async def packages(self, *, language: Language | None = None) -> list[FetchedAppPackage]:
        """Fetch the app's packages.

        Parameters
        ----------
        language
            The language to fetch the packages in. If ``None``, the current language will be used.
        """
        fetched = await self.fetch(language=language)
        return fetched._packages

    async def add_free_licenses(self) -> list[License]:
        """Request the free licenses for this app.

        Raises
        ------
        ValueError
            No licenses were granted.
        """
        info = await self._state.request_free_licenses(self.id)
        return info[self.id]

    async def shop_items(self, *, currency: Currency | None = None, language: Language | None = None) -> AppShopItems:
        """Fetch the items that are purchasable inside of the app's shop.

        Parameters
        ----------
        currency
            If passed only return the prices in this currency otherwise return the prices in all supported currencies.
        language
            The language to resolve the descriptions of the items to. If ``None`` uses the current language.
        """
        data = await self._state.http.get_app_asset_prices(self.id, currency)
        tags = (
            [
                AppShopItemTag(name, display_name, id)
                for (display_name, name), id in zip(data["tags"].items(), data["tag_ids"].values())
            ]
            if "tags" in data and "tag_ids" in data
            else []
        )

        INSTANCE_ID_0 = InstanceID(0)
        assets = {(ClassID(int(asset["classid"])), INSTANCE_ID_0): asset for asset in data["assets"]}
        return AppShopItems(
            [
                AppShopItem(self._state, assets[class_id, INSTANCE_ID_0], description, tags)
                for (class_id, _), description in (await self._state.fetch_item_info(self.id, assets, language)).items()
            ],
            tags,
        )

    async def community_items(
        self, *, type: CommunityDefinitionItemType = CommunityDefinitionItemType.NONE, language: Language | None = None
    ) -> list[CommunityItem[Self]]:
        """Fetch the app's community item definitions.

        Parameters
        ----------
        type
            The type of community item definitions to fetch.
        language
            The language to fetch the community item definitions in. If ``None``, the current language will be used.
        """
        defs = await self._state.fetch_community_item_definitions(self.id, type, language)
        return [
            CommunityItem(
                self._state,
                def_.item_type,
                self,
                def_.item_name,
                def_.item_title,
                def_.item_description,
                (
                    CDNAsset(
                        self._state,
                        f"{URL.CDN}/steamcommunity/public/images/items/{def_.item_image_large or def_.item_image_small}",
                    )
                    if def_.item_image_large or def_.item_image_small
                    else None
                ),
                (
                    CDNAsset(
                        self._state,
                        f"{URL.CDN}/steamcommunity/public/images/items/{def_.item_movie_mp4 or def_.item_movie_mp4_small}",
                    )
                    if def_.item_movie_mp4 or def_.item_movie_mp4_small
                    else None
                ),
                JSON_LOADS(def_.item_key_values) if def_.item_key_values else None,
                def_.item_series,
                CommunityItemClass.try_value(def_.item_class),
                await self._state._maybe_user(def_.editor_accountid) if def_.editor_accountid else None,
                def_.active,
                (
                    CDNAsset(
                        self._state,
                        f"https://community.cloudflare.steamstatic.com/economy/image/{def_.item_image_composed}",
                    )
                    if def_.item_image_composed
                    else None
                ),
                (
                    CDNAsset(
                        self._state, f"{URL.CDN}/steamcommunity/public/images/items/{def_.item_image_composed_foil}"
                    )
                    if def_.item_image_composed_foil
                    else None
                ),
                def_.deleted,
                DateTime.from_timestamp(def_.item_last_changed) if def_.item_last_changed else None,
            )
            for def_ in defs
        ]

    async def reward_items(self, *, language: Language | None = None) -> list[RewardItem[Self]]:
        """Fetch the app's reward item definitions.

        Parameters
        ----------
        language
            The language to fetch the reward item definitions in. If ``None``, the current language will be used.
        """
        defs = await self._state.fetch_reward_items(self.id, language)
        items = [
            RewardItem(
                def_.type,
                self,
                def_.community_item_data.item_name,
                def_.community_item_data.item_title,
                def_.community_item_data.item_description,
                def_.internal_description,  # really seems to be the name?
                (
                    CDNAsset(
                        self._state,
                        (
                            f"{URL.CDN}/steamcommunity/public/images/items/"
                            f"{def_.community_item_data.item_image_large or def_.community_item_data.item_image_small}"
                        ),
                    )
                    if def_.community_item_data.item_image_large or def_.community_item_data.item_image_small
                    else None
                ),
                (
                    CDNAsset(
                        self._state,
                        (
                            f"{URL.CDN}/steamcommunity/public/images/items/"
                            f"{def_.community_item_data.item_movie_mp4 or def_.community_item_data.item_movie_mp4_small}"
                        ),
                    )
                    if def_.community_item_data.item_movie_mp4 or def_.community_item_data.item_movie_mp4_small
                    else None
                ),
                def_.community_item_data.animated,
                tuple(
                    AppBadge(
                        self._state,
                        1,
                        f"{def_.community_item_data.item_name.removesuffix(' Point Shop Badge')} - Level {badge.level}",
                        self,
                        CDNAsset(self._state, f"{URL.CDN}/steamcommunity/public/images/items/{self.id}/{badge.image}"),
                        badge.level,
                    )
                    for badge in def_.community_item_data.badge_data
                ),
                def_.defid,
                def_.quantity,
                CommunityItemClass.try_value(def_.community_item_class),
                def_.community_item_type,
                def_.point_cost,
                DateTime.from_timestamp(def_.timestamp_created),
                DateTime.from_timestamp(def_.timestamp_updated),
                DateTime.from_timestamp(def_.timestamp_available),
                DateTime.from_timestamp(def_.timestamp_available_end),
                def_.active,
                def_.community_item_data.profile_theme_id,
                timedelta(seconds=def_.usable_duration),
                def_.bundle_discount,
            )
            for def_ in defs
        ]
        for item, def_ in zip(items, defs):
            if def_.bundle_defids:
                item.bundles = cast(
                    "tuple[RewardItem[Self], ...]", tuple(utils.get(items, def_index=id) for id in def_.bundle_defids)
                )

        return items

    async def badges(self, *, language: Language | None = None) -> Sequence[AppBadge[Self]]:
        """Fetch this app's badges.

        Parameters
        ----------
        language
            The language to fetch the badges in. If ``None``, the current language will be used.
        """
        if self.id == STEAM.id:
            # 753 doesn't have a community item definition, so we just have the badges hardcoded
            return [
                AppBadge(self._state, badge.id, badge.name, self, CDNAsset(self._state, badge.url), level=badge.level)
                for badge in STEAM_BADGES
            ]

        community_items, rewards = await asyncio.gather(
            self.community_items(language=language), self.reward_items(language=language)
        )

        badge_def = utils.get(community_items, class_=CommunityItemClass.Badge)
        badge_rewards = utils.get(rewards, class_=CommunityItemClass.Badge)

        if badge_def is None and badge_rewards is None:
            return []
        if badge_def is not None and badge_rewards is None:
            return badge_def.badges
        if badge_def is None and badge_rewards is not None:
            return badge_rewards.badges

        assert badge_def
        assert badge_rewards
        return [*badge_def.badges, *badge_rewards.badges]

    async def legacy_cd_key(self) -> str:
        """Fetch the legacy CD key for this app."""
        return await self._state.fetch_legacy_cd_key(self.id)

    async def encrypted_ticket(self, key: bytes, *, user_data: bytes = b"") -> EncryptedTicket:
        """Fetch an encrypted ticket for this app.

        Parameters
        ----------
        key
            The key to encrypt the ticket with.
        user_data
            The user data to include in the ticket.
        """
        encrypted_ticket = await self._state.fetch_encrypted_app_ticket(self.id, user_data)
        return EncryptedTicket(self._state, encrypted_ticket, key)

    async def ownership_ticket(self) -> OwnershipTicket:
        """Fetch an ownership ticket for this app."""
        ticket = await self._state.fetch_app_ownership_ticket(self.id)
        return await parse_app_ticket(self._state, utils.StructIO(ticket))

    @asynccontextmanager
    async def create_authentication_ticket(self) -> AsyncGenerator[AuthenticationTicket, None]:
        """Create an authentication ticket for this app.

        Examples
        --------

        .. code:: python

            async with app.create_authentication_ticket() as ticket:
                ...  # send the ticket to a user or server
                # ticket will only be valid inside this block
        """
        # for the ticket to be valid we have to be playing the game
        async with self._state.temporarily_play(self):
            ownership_ticket = await self._state.fetch_app_ownership_ticket(self.id)
            with utils.StructIO() as io:
                bytes = self._state._game_connect_bytes.pop(0)
                io.write_u32(len(bytes))
                io.write(bytes)
                io.write_u32(24)
                io.write_u32(1)
                io.write_u32(2)
                io.write_u32(int(self._state.ws.public_ip))
                io.write_u32(0)
                io.write_u32(int((self._state.steam_time.timestamp() - self._state.ws.connect_time.timestamp()) * 1000))
                self._state.connection_count += 1
                io.write_u32(self._state.connection_count)
                io.write_u32(len(ownership_ticket))
                io.write(ownership_ticket)
                io.seek(0)
                ticket = await parse_app_ticket(self._state, io)
                assert isinstance(ticket, AuthenticationTicket)
                try:
                    await ticket.activate()
                    yield ticket
                finally:
                    await ticket.deactivate()


class Apps(PartialApp[str], Enum):
    """This is "enum" to trick type checkers into allowing Literal[TF2] to be valid for overloads in extensions."""

    __slots__ = ("_name",)

    @classmethod
    def _new_member(cls, *, name: str, value: tuple[str, int]) -> Self:
        self = object.__new__(cls)
        set_attribute = object.__setattr__
        set_attribute(self, "_name", name)
        set_attribute(self, "name", value[0])
        set_attribute(self, "id", value[1])
        return self

    def __repr__(self) -> str:
        return self._name

    @property
    def value(self) -> int:
        return self.id

    TF2 = "Team Fortress 2", 440
    LFD2 = "Left 4 Dead 2", 550
    DOTA2 = "DOTA 2", 570
    CSGO = "Counter Strike Global-Offensive", 730
    STEAM = "Steam", 753

    @property
    def _state(self) -> ConnectionState:
        try:
            return STATE.get()
        except LookupError:
            raise ValueError("Cannot access the state of constant apps outside of a client.") from None


TF2 = Apps.TF2
"""The Team Fortress 2 app."""
DOTA2 = Apps.DOTA2
"""The DOTA 2 app."""
CSGO = Apps.CSGO
"""The Counter Strike Global-Offensive app."""
LFD2 = Apps.LFD2
"""The Left 4 Dead 2 app."""
STEAM = Apps.STEAM
"""The Steam app with context ID 6 (gifts)."""


class OwnershipDLC(PartialApp):
    def __init__(self, state: ConnectionState, *, id: int, owned_packages: list[PartialPackage]):
        super().__init__(state, id=id)
        self.owned_packages = owned_packages
        """The packages that the ticket owner owns which grant them this DLC."""


@dataclass(slots=True)
class PartialAppPriceOverview:
    currency: Currency
    initial: int
    final: int
    discount_percent: int


class DLC(PartialApp[str]):
    """Represents DLC (downloadable content) for an app."""

    __slots__ = (
        "created_at",
        "logo",
        "price_overview",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
    )

    def __init__(self, state: ConnectionState, data: app.DLC):
        super().__init__(state, id=data["id"], name=data["name"])
        self.created_at = (
            DateTime.from_timestamp(int(data["release_date"]["steam"])) if data["release_date"]["steam"] else None
        )
        """The time the DLC was released at."""
        self.logo = CDNAsset(state, data["header_image"])
        """The logo url of the DLC."""
        currency_code = Currency[data["price_overview"].pop("currency")]  # type: ignore  # TypedDict is immutable
        self.price_overview = PartialAppPriceOverview(currency=currency_code, **data["price_overview"])  # type: ignore
        """A price overview for the DLC."""

        platforms = data["platforms"]
        self._on_windows = platforms["windows"]
        self._on_mac_os = platforms["mac"]
        self._on_linux = platforms["linux"]

    def is_free(self) -> bool:
        """Whether the app is free to download."""
        return not self.price_overview.final

    def is_on_windows(self) -> bool:
        """Whether the app is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the app is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the app is playable on Linux."""
        return self._on_linux


class UserApp(PartialApp[str]):
    """Represents a Steam app fetched by :meth:`steam.User.apps`."""

    __slots__ = (
        "playtime_forever",
        "playtime_two_weeks",
        "playtime_windows",
        "playtime_mac_os",
        "playtime_linux",
        "icon",
        "last_played_at",
        "content_descriptors",
        "_stats_visible",
        "_workshop",
        "_market",
        "_dlc",
        "_leaderboards",
    )

    def __init__(self, state: ConnectionState, proto: player.GetOwnedGamesResponseGame):
        super().__init__(state=state, id=proto.appid, name=proto.name)
        self.playtime_forever: timedelta = timedelta(minutes=proto.playtime_forever)
        """The total time the app has been played for."""
        self.playtime_two_weeks: timedelta = timedelta(minutes=proto.playtime_2_weeks)
        """The amount of time the user has played the app in the last two weeks."""
        self.playtime_windows: timedelta = timedelta(minutes=proto.playtime_windows_forever)
        """The total amount of time the user has played the app on Windows."""
        self.playtime_mac_os: timedelta = timedelta(minutes=proto.playtime_mac_forever)
        """The total amount of time the user has played the app on macOS."""
        self.playtime_linux: timedelta = timedelta(minutes=proto.playtime_linux_forever)
        """The total amount of time the user has played the app on Linux."""
        self.icon = CDNAsset(
            state, str(URL.CDN / f"steamcommunity/public/images/apps/{self.id}/{proto.img_icon_url}.jpg")
        )
        """The icon of the app."""
        self.last_played_at = DateTime.from_timestamp(proto.rtime_last_played)
        """The time the user last played this app at."""

        self.content_descriptors = [
            ContentDescriptor.try_value(descriptor) for descriptor in proto.content_descriptorids
        ]
        """The content descriptors of the app."""

        self._stats_visible = proto.has_community_visible_stats
        self._workshop = proto.has_workshop
        self._market = proto.has_market
        self._dlc = proto.has_dlc
        self._leaderboards = proto.has_leaderboards

    def has_visible_stats(self) -> bool:
        """Whether the app has publicly visible stats."""
        return self._stats_visible

    def has_workshop(self) -> bool:
        """Whether the app has a workshop."""
        return self._workshop

    def has_market(self) -> bool:
        """Whether the app has a market."""
        return self._market

    def has_dlc(self) -> bool:
        """Whether the app has DLC."""
        return self._dlc

    def has_leaderboards(self) -> bool:
        """Whether the app has leaderboards."""
        return self._leaderboards


class UserRecentlyPlayedApp(PartialApp[str]):
    def __init__(self, state: ConnectionState, data: app.UserRecentlyPlayedApp):
        super().__init__(state, id=data["appid"], name=data["name"])
        self.playtime_forever: timedelta = timedelta(minutes=data["playtime_forever"])
        """The total time the app has been played for."""
        self.playtime_two_weeks: timedelta = timedelta(minutes=data["playtime_2weeks"])
        """The amount of time the user has played the app in the last two weeks."""
        self.icon = CDNAsset(
            state, str(URL.CDN / f"steamcommunity/public/images/apps/{self.id}/{data['img_icon_url']}.jpg")
        )
        """The icon of the app."""


class WishlistApp(PartialApp[str]):
    """Represents a Steam app fetched by :meth:`steam.User.wishlist`\\."""

    __slots__ = (
        "priority",
        "added_at",
        "background",
        "created_at",
        "logo",
        "rank",
        "review_status",
        "score",
        "screenshots",
        # "tags",
        "total_reviews",
        "type",
        "partial_packages",
        "partial_bundles",
        "_free",
        "_on_linux",
        "_on_mac_os",
        "_on_windows",
    )

    def __init__(self, state: ConnectionState, id: int, data: app.WishlistApp):
        from .bundle import PartialBundle
        from .package import PartialPackage

        super().__init__(state, id=id, name=data["name"])
        self.priority = int(data["priority"])
        """The priority of the app in the wishlist."""
        self.type = AppType.from_str(data["type"])
        """The type of the app."""
        self.logo = CDNAsset(state, data["capsule"])
        """The logo of the app."""
        self.score = data["review_score"]
        """The score of the app out of ten."""
        self.total_reviews = int(data["reviews_total"].replace(",", ""))
        """The total number reviews for the app."""
        self.review_status = ReviewType.try_value(data["review_score"])
        """The review status of the app."""
        self.created_at = DateTime.from_timestamp(float(data["release_date"])) if data["release_date"] else None
        """The time the app was uploaded at."""
        self.screenshots = [
            CDNAsset(state, f"{URL.CDN}/steam/apps/{self.id}/{screenshot_url}")
            for screenshot_url in data["screenshots"]
        ]
        """The screenshots of the app."""
        self.added_at = DateTime.from_timestamp(data["added"])
        """The time that the app was added to their wishlist."""
        self.background = CDNAsset(state, data["background"])
        """The background of the app."""
        # self.tags = [Tag(state, int(tag)) for tag in data["tags"]]
        # """The tags of the app."""
        self.rank = data["rank"]
        """The global rank of the app by popularity."""
        self.partial_packages: list[PartialPackage] = []
        """The packages this app is included in."""
        self.partial_bundles: list[PartialBundle] = []
        """The bundles this app is included in."""
        for sub in data["subs"]:
            if package_id := sub["packageid"]:
                self.partial_packages.append(PartialPackage(state, id=package_id))
            elif bundle_id := sub["bundleid"]:
                self.partial_bundles.append(PartialBundle(state, id=bundle_id))

        self._free = data["is_free_game"]
        self._on_windows = bool(data.get("win", False))
        self._on_mac_os = bool(data.get("mac", False))
        self._on_linux = bool(data.get("linux", False))

    def is_free(self) -> bool:
        """Whether the app is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the app is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the app is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the app is playable on Linux."""
        return self._on_linux


class FetchedAppMovie(_IOMixin):
    __slots__ = ("name", "id", "url", "created_at", "_state")

    def __init__(self, state: ConnectionState, movie: dict[str, Any]):
        self._state = state
        self.name: str = movie["name"]
        self.id: int = movie["id"]
        self.url: ReadOnly[str] = movie["mp4"]["max"]
        match = re.search(r"t=(\d+)", self.url)  # type: ignore  # should become unnecessary at some point
        self.created_at = DateTime.from_timestamp(int(match[1])) if match else None

    def __repr__(self) -> str:
        attrs = ("name", "id", "url", "created_at")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


@dataclass(slots=True)
class AppPriceOverview(PartialAppPriceOverview):
    initial_formatted: str
    final_formatted: str


class FetchedApp(PartialApp[str]):
    """Represents a Steam app fetched by :meth:`steam.Client.fetch_app`\\."""

    __slots__ = (
        "logo",
        "background",
        "created_at",
        "type",
        "categories",
        "partial_dlc",
        "_packages",
        "website_url",
        "developers",
        "publishers",
        "description",
        "full_description",
        "movies",
        "price_overview",
        "content_descriptors",
        "_free",
        "_on_windows",
        "_on_mac_os",
        "_on_linux",
        "_language",
    )

    def __init__(self, state: ConnectionState, data: app.FetchedApp, language: Language):
        from .package import PartialPackage

        super().__init__(state, id=data["steam_appid"], name=data["name"])
        self.logo = CDNAsset(state, data["header_image"])
        """The logo of the app."""
        self.background = CDNAsset(state, data["background"])
        """The background of the app."""
        self.created_at = (
            DateTime.parse_steam_date(data["release_date"]["date"], full_month=False)
            if data["release_date"]["date"]
            else None
        )
        """The time the app was uploaded at."""
        self.type = AppType.from_str(data["type"])
        """The type of the app."""

        self.categories = [
            Category(state, id=category["id"], name=category["description"]) for category in data["categories"]
        ]

        try:
            currency = Currency[data["price_overview"].pop("currency")]  # type: ignore
            self.price_overview = AppPriceOverview(currency=currency, **data["price_overview"])  # type: ignore
            """The price overview of the app."""
        except KeyError:
            self.price_overview = None

        self.partial_dlc = [PartialApp(state, id=dlc_id) for dlc_id in data.get("dlc", [])]
        """The app's downloadable content."""

        from .package import FetchedAppPackage

        self._packages = [
            FetchedAppPackage(state, package)
            for package_group in data["package_groups"]
            for package in package_group["subs"]
        ]
        current_package_ids = {package.id for package in self._packages}
        self._packages += [
            PartialPackage(state, id=package_id)
            for package_id in data.get("packages", [])
            if package_id not in current_package_ids
        ]

        self.website_url = data.get("website")
        """The website URL of the app."""
        self.developers = data["developers"]
        """The developers of the app."""
        self.publishers = data["publishers"]
        """The publishers of the app."""
        self.description = data["short_description"]
        """The short description of the app."""
        self.full_description = data["detailed_description"]
        """The full description of the app."""

        self.movies = [FetchedAppMovie(state, movie) for movie in data["movies"]] if "movies" in data else None
        """
        A list of the app's movies, each of which has ``name``\\, ``id``\\, ``url`` and optional ``created_at``
        attributes.
        """
        self.content_descriptors = [
            ContentDescriptor.try_value(descriptor) for descriptor in data.get("content_descriptors", {}).get("ids", [])
        ]
        """The app's content descriptors for explicit content"""

        self._free = data["is_free"]
        self._on_windows = bool(data["platforms"].get("windows", False))
        self._on_mac_os = bool(data["platforms"].get("mac", False))
        self._on_linux = bool(data["platforms"].get("linux", False))

        self._language = language

    async def packages(self, *, language: Language | None = None) -> list[FetchedAppPackage]:
        if language is not self._language:
            return await super().packages(language=language)
        return self._packages

    def is_free(self) -> bool:
        """Whether the app is free to download."""
        return self._free

    def is_on_windows(self) -> bool:
        """Whether the app is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the app is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the app is playable on Linux."""
        return self._on_linux


@dataclass(slots=True)
class UserInventoryInfoContext:
    """Represents a context ID type."""

    id: ContextID
    """The ID of the context ID type."""
    name: str
    """The name of the context ID type."""
    count: int
    """The number of items in the context ID type."""


class UserInventoryInfoApp(PartialApp[str]):
    __slots__ = ("icon", "inventory_logo")

    def __init__(self, state: ConnectionState, id: int, name: str, icon_url: str, inventory_logo_url: str):
        super().__init__(state, id=id, name=name)
        self.icon = CDNAsset(state, icon_url)
        """The icon of the app."""
        self.inventory_logo = CDNAsset(state, inventory_logo_url)
        """The inventory logo of the app."""


class AppListApp(PartialApp[str]):
    """Represents an app returned by :class:`~steam.Client.all_apps`."""

    __slots__ = ("last_modified", "price_change_number")

    def __init__(self, state: ConnectionState, data: app.AppListApp):
        super().__init__(state, id=data["appid"], name=data["name"])
        self.last_modified = DateTime.from_timestamp(data["last_modified"])
        """The time the app was last modified at."""
        self.price_change_number = data["price_change_number"]
        """The price change number of the app."""



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_parental.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


@dataclass(eq=False, repr=False)
class ParentalApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    is_allowed: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class ParentalSettings(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    applist_base_id: int = betterproto.uint32_field(2)
    applist_base_description: str = betterproto.string_field(3)
    applist_base: "list[ParentalApp]" = betterproto.message_field(4)
    applist_custom: "list[ParentalApp]" = betterproto.message_field(5)
    passwordhashtype: int = betterproto.uint32_field(6)
    salt: bytes = betterproto.bytes_field(7)
    passwordhash: bytes = betterproto.bytes_field(8)
    is_enabled: bool = betterproto.bool_field(9)
    enabled_features: int = betterproto.uint32_field(10)
    recovery_email: str = betterproto.string_field(11)
    is_site_license_lock: bool = betterproto.bool_field(12)


class EnableParentalSettingsRequest(UnifiedMessage, um_name="Parental.EnableParentalSettings"):
    password: str = betterproto.string_field(1)
    settings: "ParentalSettings" = betterproto.message_field(2)
    sessionid: str = betterproto.string_field(3)
    enablecode: int = betterproto.uint32_field(4)
    steamid: int = betterproto.fixed64_field(10)


class EnableParentalSettingsResponse(UnifiedMessage, um_name="Parental.EnableParentalSettings"):
    pass


class DisableParentalSettingsRequest(UnifiedMessage, um_name="Parental.DisableParentalSettings"):
    password: str = betterproto.string_field(1)
    steamid: int = betterproto.fixed64_field(10)


class DisableParentalSettingsResponse(UnifiedMessage, um_name="Parental.DisableParentalSettings"):
    pass


class GetParentalSettingsRequest(UnifiedMessage, um_name="Parental.GetParentalSettings"):
    steamid: int = betterproto.fixed64_field(10)


class GetParentalSettingsResponse(UnifiedMessage, um_name="Parental.GetParentalSettings"):
    settings: "ParentalSettings" = betterproto.message_field(1)


class GetSignedParentalSettingsRequest(UnifiedMessage, um_name="Parental.GetSignedParentalSettings"):
    priority: int = betterproto.uint32_field(1)


class GetSignedParentalSettingsResponse(UnifiedMessage, um_name="Parental.GetSignedParentalSettings"):
    serialized_settings: bytes = betterproto.bytes_field(1)
    signature: bytes = betterproto.bytes_field(2)


class SetParentalSettingsRequest(UnifiedMessage, um_name="Parental.SetParentalSettings"):
    password: str = betterproto.string_field(1)
    settings: "ParentalSettings" = betterproto.message_field(2)
    new_password: str = betterproto.string_field(3)
    sessionid: str = betterproto.string_field(4)
    steamid: int = betterproto.fixed64_field(10)


class SetParentalSettingsResponse(UnifiedMessage, um_name="Parental.SetParentalSettings"):
    pass


class ValidateTokenRequest(UnifiedMessage, um_name="Parental.ValidateToken"):
    unlock_token: str = betterproto.string_field(1)


class ValidateTokenResponse(UnifiedMessage, um_name="Parental.ValidateToken"):
    pass


class ValidatePasswordRequest(UnifiedMessage, um_name="Parental.ValidatePassword"):
    password: str = betterproto.string_field(1)
    session: str = betterproto.string_field(2)
    send_unlock_on_success: bool = betterproto.bool_field(3)


class ValidatePasswordResponse(UnifiedMessage, um_name="Parental.ValidatePassword"):
    token: str = betterproto.string_field(1)


class LockClientRequest(UnifiedMessage, um_name="Parental.LockClient"):
    session: str = betterproto.string_field(1)


class LockClientResponse(UnifiedMessage, um_name="Parental.LockClient"):
    pass


class RequestRecoveryCodeRequest(UnifiedMessage, um_name="Parental.RequestRecoveryCode"):
    pass


class RequestRecoveryCodeResponse(UnifiedMessage, um_name="Parental.RequestRecoveryCode"):
    pass


class DisableWithRecoveryCodeRequest(UnifiedMessage, um_name="Parental.DisableWithRecoveryCode"):
    recovery_code: int = betterproto.uint32_field(1)
    steamid: int = betterproto.fixed64_field(10)


class DisableWithRecoveryCodeResponse(UnifiedMessage, um_name="Parental.DisableWithRecoveryCode"):
    pass


@dataclass(eq=False, repr=False)
class ParentalSettingsChangeNotification(betterproto.Message):
    serialized_settings: bytes = betterproto.bytes_field(1)
    signature: bytes = betterproto.bytes_field(2)
    password: str = betterproto.string_field(3)
    sessionid: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ParentalUnlockNotification(betterproto.Message):
    password: str = betterproto.string_field(1)
    sessionid: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ParentalLockNotification(betterproto.Message):
    sessionid: str = betterproto.string_field(1)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import Callable, Generator, Mapping
from enum import Enum as _Enum, EnumMeta as _EnumMeta
from functools import reduce
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Final, Generic, Literal, SupportsInt, TypeVar, cast

from ._const import DOCS_BUILDING

if TYPE_CHECKING:
    from typing_extensions import Never, Self

__all__ = (
    "Enum",
    "IntEnum",
    "Flags",
    "Intents",
    "Result",
    "Language",
    "Currency",
    "Realm",
    "PurchaseResult",
    "Universe",
    "Type",
    "TypeChar",
    "Instance",
    "FriendRelationship",
    "PersonaState",
    "PersonaStateFlag",
    "CommunityVisibilityState",
    "TradeOfferState",
    "ChatMemberRank",
    "ChatMemberJoinState",
    "ChatEntryType",
    "ClanAccountFlags",
    "UIMode",
    "ReviewType",
    "GameServerRegion",
    "EventType",
    "ProfileItemType",
    "ProfileCustomisationStyle",
    "CommunityItemClass",
    "ProfileItemEquippedFlag",
    "DepotFileFlag",
    "AppType",
    "LicenseFlag",
    "LicenseType",
    "BillingType",
    "PaymentMethod",
    "PackageStatus",
    "PublishedFileRevision",
    "PublishedFileType",
    "PublishedFileVisibility",
    "PublishedFileQueryType",
    "PublishedFileQueryFileType",
    "LeaderboardUploadScoreMethod",
    "LeaderboardSortMethod",
    "LeaderboardDisplayType",
    "LeaderboardDataRequest",
    "CommunityDefinitionItemType",
    "AuthSessionResponse",
    "ContentDescriptor",
    "UserNewsType",
)

T = TypeVar("T")

T_co = TypeVar("T_co", covariant=True)
TT = TypeVar("TT", bound="type[Any]")


class classproperty(Generic[TT, T_co]):
    def __init__(self, func: Callable[[TT], T_co]):
        self.__func__ = func

    def __get__(self, instance: Any, type: TT) -> T_co:
        return self.__func__(type)


def _is_descriptor(obj: object, /) -> bool:
    """Returns True if obj is a descriptor, False otherwise."""
    return hasattr(obj, "__get__") or hasattr(obj, "__set__") or hasattr(obj, "__delete__")


class EnumDict(dict[str, Any]):
    """Required to detect the difference between:

    class MyEnum(steam.Enum):
        A = 1
        B = 2
        C = A  # this is an alias for A and not a new member
        D = 2  # this is a distinct member
    """

    def __init__(self):
        self.aliases: set[str] = set()

    def __getitem__(self, key: str) -> Any:
        self.aliases.add(key)
        return super().__getitem__(key)


class EnumType(_EnumMeta if TYPE_CHECKING else type):
    _value_map_: Mapping[Any, Enum]
    _member_map_: Mapping[str, Enum]  # type: ignore

    @classmethod
    def __prepare__(mcs, name: str, bases: tuple[type, ...]) -> EnumDict:  # type: ignore
        return EnumDict()

    def __new__(mcs, name: str, bases: tuple[type, ...], namespace: EnumDict) -> type[Enum]:
        value_map: dict[Any, Enum] = {}
        member_map: dict[str, Enum] = {}

        new_mcs: type[EnumType] = type(
            f"{name}Type",
            tuple(
                dict.fromkeys([base.__class__ for base in bases if base.__class__ is not type] + [EnumType, type])
            ),  # reorder the bases so EnumType and type are last to avoid conflicts
            {"_value_map_": value_map, "_member_map_": member_map},
        )  # type: ignore

        members = {name: value for name, value in namespace.items() if not _is_descriptor(value) and name[0] != "_"}

        cls = cast(
            "type[Enum]",
            type.__new__(new_mcs, name, bases, {key: value for key, value in namespace.items() if key not in members}),
        )  # this allows us to disallow member access from other members as members become proper class variables

        for name, value in members.items():
            if (member := value_map.get(value)) is None or member.name not in namespace.aliases:
                member = cls._new_member(name=name, value=value)
                value_map[value] = member

            member_map[name] = member
            type.__setattr__(new_mcs, name, member)

        return cls

    if not TYPE_CHECKING:

        def __iter__(cls) -> Generator[Enum, None, None]:
            yield from cls._member_map_.values()

        def __reversed__(cls) -> Generator[Enum, None, None]:
            yield from reversed(cls._member_map_.values())

        def __getitem__(cls, key: str) -> Enum:
            return cls._member_map_[key]

        @property
        def __members__(cls) -> MappingProxyType[str, Enum]:
            return MappingProxyType(cls._member_map_)

    def __repr__(cls) -> str:
        return f"<enum {cls.__name__!r}>"

    def __len__(cls) -> int:
        return len(cls._member_map_)

    def __setattr__(cls, name: str, value: Any) -> Never:
        if name.startswith("__") and name.endswith("__"):
            return super().__setattr__(name, value)  # type: ignore
        raise AttributeError(f"{cls.__name__}: cannot reassign Enum members.")

    def __delattr__(cls, name: str) -> Never:
        raise AttributeError(f"{cls.__name__}: cannot delete Enum members.")

    def __contains__(cls, member: object) -> bool:
        return isinstance(member, Enum) and isinstance(member, cls) and member.name in cls._member_map_

    def __dir__(self) -> list[str]:
        return super().__dir__() + list(self._member_map_)


# pretending these are enum subclasses makes things much nicer for linters as enums have custom behaviour you can't
# replicate in the current type system
class Enum(_Enum if TYPE_CHECKING else object, metaclass=EnumType):
    """A general enumeration, emulates `enum.Enum`."""

    _member_map_: Mapping[str, Self]  # type: ignore  # yes this is unsafe
    _value_map_: Mapping[Any, Self]

    def __new__(cls, value: Any) -> Self:
        try:
            return cls._value_map_[value]
        except (KeyError, TypeError):
            raise ValueError(f"{value!r} is not a valid {cls.__name__}") from None

    @classmethod
    def _new_member(cls, *, name: str, value: Any) -> Self:
        self = (
            super().__new__(cls, value)
            if any(not issubclass(base, Enum) for base in cls.__mro__[:-1])  # is it is a mixin enum
            else super().__new__(cls)  # type: ignore
        )
        super().__setattr__(self, "name", name)
        super().__setattr__(self, "value", value)

        return self

    if not DOCS_BUILDING:

        def __setattr__(self, key: str, value: Any) -> Never:
            raise AttributeError(f"Cannot reassign {self.__class__.__name__} members attribute's.")

        def __delattr__(self, item: Any) -> Never:
            raise AttributeError(f"Cannot delete {self.__class__.__name__} attribute's.")

    def __bool__(self) -> Literal[True]:
        return True  # an enum member with a zero value would return False otherwise

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"

    @classmethod
    def try_value(cls, value: Any, /) -> Self:
        try:
            return cls._value_map_[value]
        except (KeyError, TypeError):
            return cls._new_member(name=f"{cls.__name__}UnknownValue", value=value)


class IntEnum(Enum, int):
    """An enumeration where all the values are integers, emulates `enum.IntEnum`."""

    if TYPE_CHECKING:

        def __new__(cls, value: int) -> Self: ...

        @classmethod
        def try_value(cls, value: int) -> Self: ...


class Flags(IntEnum):
    @classmethod
    def try_value(cls, value: int) -> Self:
        if value == 0:
            # causes flags to be iter(cls) does |= on every value (which is not 0) it always returns UnknownValue
            return super().try_value(value)
        flags: Generator[Self, None, None] = (enum for enum in cls if enum.value & value)  # type: ignore
        returning_flag = next(flags, None)
        if returning_flag is not None:
            for flag in flags:
                returning_flag |= flag
            if returning_flag == value:
                return returning_flag
        return cls._new_member(name=f"{cls.__name__}UnknownValue", value=value)

    def __or__(self, other: SupportsInt) -> Self:
        cls = self.__class__
        value = self.value | int(other)
        try:
            return cls._value_map_[value]
        except KeyError:
            return cls._new_member(name=f"{self.name} | {getattr(other, 'name', other)}", value=value)

    def __and__(self, other: SupportsInt) -> Self:
        cls = self.__class__
        value = self.value & int(other)
        try:
            return cls._value_map_[value]
        except KeyError:
            return cls._new_member(name=f"{self.name} & {getattr(other, 'name', other)}", value=value)

    def __invert__(self) -> Self:
        member = self.try_value(~-self.value)
        if member.name.endswith("UnknownValue"):
            object.__setattr__(member, "name", f"~{self.name}")
        return member

    __ror__ = __or__
    __rand__ = __and__


# fmt: off
class Intents(Flags):
    """Flags that control what features the client will interact with."""
    NONE        = 0
    """Very little interaction happens from our end of the connection."""
    Users       = 1 << 0
    """Users and members are cached."""
    TradeOffers = 1 << 1
    """TradeOffers are fetched and processed automatically"""
    Messages    = 1 << 2
    """Messages are cached, cannot be enabled without :attr:`Chat`."""
    Chat        = 1 << 3
    """Chats/Channels are cached."""
    ChatGroups  = 1 << 4
    """Groups/Clans are cached."""
    Market      = 1 << 5
    """Market interactions are enabled.

    Warning
    -------
    Using this comes could lead to your account being banned, much more so than with any other library operations.
    """

    @classproperty
    def All(cls: type[Self]) -> Self:  # type: ignore
        """Returns all the intents.

        Warning
        -------
        This will return unsafe intents. Use at your own risk!

        See also
        --------
        :attr:`Safe` as a safe alternative to this function which is less likely to get you banned.
        """
        return reduce(cls.__or__, cls._value_map_.values())

    @classproperty
    def Safe(cls: type[Self]) -> Self:  # type: ignore
        """Returns all the intents without the unsafe ones."""
        return cls.All & ~cls.Market


class Result(IntEnum):
    """The result of a Steam API call. Read more on :works:`steamworks <api/steam_api#EResult>`."""
    # these are a combination of https://partner.steamgames.com/doc/api/steam_api#EResult and https://steamerrors.com
    Invalid                         = 0
    """Invalid Result."""
    OK                              = 1
    """Success."""
    Fail                            = 2
    """Generic failure."""
    NoConnection                    = 3
    """Your Steam client doesn't have a connection to the back-end."""
    InvalidPassword                 = 5
    """Password/ticket is invalid."""
    LoggedInElsewhere               = 6
    """Same user logged in elsewhere."""
    InvalidProtocolVersion          = 7
    """Protocol version is incorrect."""
    InvalidParameter                = 8
    """A parameter is incorrect."""
    FileNotFound                    = 9
    """File was not found."""
    Busy                            = 10
    """Called method busy - action not taken."""
    InvalidState                    = 11
    """Called object was in an invalid state."""
    InvalidName                     = 12
    """The name was invalid."""
    InvalidEmail                    = 13
    """The email was invalid."""
    DuplicateName                   = 14
    """The name is not unique."""
    AccessDenied                    = 15
    """Access is denied."""
    Timeout                         = 16
    """Operation timed out."""
    Banned                          = 17
    """VAC2 banned."""
    AccountNotFound                 = 18
    """Account not found."""
    InvalidSteamID                  = 19
    """The Steam ID was invalid."""
    ServiceUnavailable              = 20
    """The requested service is currently unavailable."""
    NotLoggedOn                     = 21
    """The user is not logged on."""
    Pending                         = 22
    """Request is pending (may be in process, or waiting on third party)."""
    EncryptionFailure               = 23
    """Encryption or decryption failed."""
    InsufficientPrivilege           = 24
    """Insufficient privilege."""
    LimitExceeded                   = 25
    """Too much of a good thing."""
    Revoked                         = 26
    """Access has been revoked (used for revoked guest passes)."""
    Expired                         = 27
    """License/Guest pass the user is trying to access is expired."""
    AlreadyRedeemed                 = 28
    """Guest pass has already been redeemed by account, cannot be acknowledged again."""
    DuplicateRequest                = 29
    """The request is a duplicate, ignored this time."""
    AlreadyOwned                    = 30
    """All the games in guest pass redemption request are already owned by the user."""
    IPNotFound                      = 31
    """IP address not found."""
    PersistFailed                   = 32
    """Failed to write change to the data store."""
    LockingFailed                   = 33
    """Failed to acquire access lock for this operation."""
    LogonSessionReplaced            = 34
    """The logon session has been replaced."""
    ConnectFailed                   = 35
    """Failed to connect."""
    HandshakeFailed                 = 36
    """The authentication handshake has failed."""
    IOFailure                       = 37
    """Generic IO failure."""
    RemoteDisconnect                = 38
    """The remote server has disconnected."""
    ShoppingCartNotFound            = 39
    """Failed to find the shopping cart requested."""
    Blocked                         = 40
    """A user blocked the action."""
    Ignored                         = 41
    """The target is ignoring sender."""
    NoMatch                         = 42
    """Nothing matching the request found."""
    AccountDisabled                 = 43
    """The account is disabled."""
    ServiceReadOnly                 = 44
    """This service is not accepting content changes right now."""
    AccountNotFeatured              = 45
    """Account doesn't have value, so this feature isn't available."""
    AdministratorOK                 = 46
    """Allowed to take this action, but only because requester is admin."""
    ContentVersion                  = 47
    """A Version mismatch in content transmitted within the Steam protocol."""
    TryAnotherCM                    = 48
    """The current CM can't service the user making a request, should try another."""
    PasswordRequiredToKickSession   = 49
    """You are already logged in elsewhere, this cached credential login has failed."""
    AlreadyLoggedInElsewhere        = 50
    """You are already logged in elsewhere, you must wait."""
    Suspended                       = 51
    """Long running operation (content download) suspended/paused."""
    Cancelled                       = 52
    """Operation canceled (typically by user content download)."""
    DataCorruption                  = 53
    """Operation canceled because data is malformed or unrecoverable."""
    DiskFull                        = 54
    """Operation canceled - not enough disk space."""
    RemoteCallFailed                = 55
    """An remote call or IPC call failed."""
    ExternalAccountUnlinked         = 57
    """External account is not linked to a Steam account."""
    PSNTicketInvalid                = 58
    """PSN ticket was invalid."""
    ExternalAccountAlreadyLinked    = 59
    """External account is already linked to some other account."""
    RemoteFileConflict              = 60
    """The sync cannot resume due to a conflict between the local and remote files."""
    IllegalPassword                 = 61
    """The requested new password is not legal."""
    SameAsPreviousValue             = 62
    """New value is the same as the old one (secret question and answer)."""
    AccountLogonDenied              = 63
    """Account login denied due to 2nd factor authentication failure."""
    CannotUseOldPassword            = 64
    """The requested new password is not legal."""
    InvalidLoginAuthCode            = 65
    """Account login denied due to auth code invalid."""
    AccountLogonDeniedNoMail        = 66
    """Account login denied due to 2nd factor authentication failure."""
    HardwareNotCapableOfIPT         = 67
    """The user's hardware does not support Intel's identity protection technology."""
    IPTInitError                    = 68
    """Intel's Identity Protection Technology has failed to initialize."""
    ParentalControlRestricted       = 69
    """Operation failed due to parental control restrictions for current user."""
    FacebookQueryError              = 70
    """Facebook query returned an error."""
    ExpiredLoginAuthCode            = 71
    """Account login denied due to auth code expired."""
    IPLoginRestrictionFailed        = 72
    """The login failed due to an IP restriction."""
    AccountLockedDown               = 73
    """The current users account is currently locked for use."""
    VerifiedEmailRequired           = 74
    """The logon failed because the accounts email is not verified."""
    NoMatchingURL                   = 75
    """There is no url matching the provided values."""
    BadResponse                     = 76
    """Parse failure, missing field, etc."""
    RequirePasswordReEntry          = 77
    """The user cannot complete the action until they re-enter their password."""
    ValueOutOfRange                 = 78
    """The value entered is outside the acceptable range."""
    UnexpectedError                 = 79
    """Something happened that we didn't expect to ever happen."""
    Disabled                        = 80
    """The requested service has been configured to be unavailable."""
    InvalidCEGSubmission            = 81
    """The set of files submitted to the CEG server are not valid."""
    RestrictedDevice                = 82
    """The device being used is not allowed to perform this action."""
    RegionLocked                    = 83
    """The action could not be complete because it is region restricted."""
    RateLimitExceeded               = 84
    """Temporary rate limit exceeded. Different from :attr:`LimitExceeded`."""
    LoginDeniedNeedTwoFactor        = 85
    """Need two-factor code to log in."""
    ItemDeleted                     = 86
    """The thing we're trying to access has been deleted."""
    AccountLoginDeniedThrottle      = 87
    """Login attempt failed, try to throttle response to possible attacker."""
    TwoFactorCodeMismatch           = 88
    """Two-factor code mismatch."""
    TwoFactorActivationCodeMismatch = 89
    """Activation code for two-factor didn't match."""
    NotModified                     = 91
    """Data not modified."""
    TimeNotSynced                   = 93
    """The time presented is out of range or tolerance."""
    SMSCodeFailed                   = 94
    """SMS code failure (no match, none pending, etc.)."""
    AccountActivityLimitExceeded    = 96
    """Too many changes to this account."""
    PhoneActivityLimitExceeded      = 97
    """Too many changes to this phone."""
    RefundToWallet                  = 98
    """Cannot refund to payment method, must use wallet."""
    EmailSendFailure                = 99
    """Cannot send an email."""
    NotSettled                      = 100
    """Can't perform operation till payment has settled."""
    NeedCaptcha                     = 101
    """Needs to provide a valid captcha."""
    GSLTDenied                      = 102
    """A game server login token owned by this token's owner has been banned."""
    GSOwnerDenied                   = 103
    """Game server owner is denied for other reason."""
    InvalidItemType                 = 104
    """The type of thing we were requested to act on is invalid."""
    IPBanned                        = 105
    """The IP address has been banned from taking this action."""
    GSLTExpired                     = 106
    """This Game Server Login Token has expired from disuse; can be reset for use."""
    InsufficientFunds               = 107
    """User doesn't have enough wallet funds to complete the action."""
    TooManyPending                  = 108
    """There are too many of this thing pending already."""
    NoSiteLicensesFound             = 109
    """No site licenses found."""
    WGNetworkSendExceeded           = 110
    """The WG couldn't send a response because we exceeded max network send size."""
    AccountNotFriends               = 111
    """Not friends with the relevant account."""
    LimitedUserAccount              = 112
    """The account is limited and cannot perform this action."""
    CantRemoveItem                  = 113
    """Cannot remove the item."""
    AccountHasBeenDeleted           = 114
    """The relevant account has been deleted."""
    AccountHasCancelledLicense      = 115
    """The user has a user cancelled license."""
    DeniedDueToCommunityCooldown    = 116
    """The request was denied due to community cooldown."""
    NoLauncherSpecified             = 117
    """No launcher was specified."""
    MustAgreeToSSA                  = 118
    """User must agree to China SSA or global SSA before login."""
    LauncherMigrated                = 119
    """The specified launcher type is no longer supported."""
    SteamRealmMismatch              = 120
    """The user's realm does not match the realm of the requested resource."""
    InvalidSignature                = 121
    """Signature check did not match."""
    ParseFailure                    = 122
    """Failed to parse input."""
    NoVerifiedPhone                 = 123
    """Account does not have a verified phone number."""
    InsufficientBatteryCharge       = 124
    """The device battery is too low to complete the action."""


class Language(IntEnum):
    """The language for a request."""
    NONE                        = -1
    """No language specified."""
    English                     = 0
    """English."""
    German                      = 1
    """German."""
    French                      = 2
    """French."""
    Italian                     = 3
    """Italian."""
    Korean                      = 4
    """Korean."""
    Spanish                     = 5
    """Spanish."""
    SimplifiedChinese           = 6
    """Simplified Chinese."""
    TraditionalChinese          = 7
    """Traditional Chinese."""
    Russian                     = 8
    """Russian."""
    Thai                        = 9
    """Thai."""
    Japanese                    = 10
    """Japanese."""
    Portuguese                  = 11
    """Portuguese."""
    Polish                      = 12
    """Polish."""
    Danish                      = 13
    """Danish."""
    Dutch                       = 14
    """Dutch."""
    Finnish                     = 15
    """Finnish."""
    Norwegian                   = 16
    """Norwegian."""
    Swedish                     = 17
    """Swedish."""
    Romanian                    = 18
    """Romanian."""
    Turkish                     = 19
    """Turkish."""
    Hungarian                   = 20
    """Hungarian."""
    Czech                       = 21
    """Czech."""
    PortugueseBrazil            = 22
    """Brazilian Portuguese."""
    Bulgarian                   = 23
    """Bulgarian."""
    Greek                       = 24
    """Greek."""
    Arabic                      = 25
    """Arabic."""
    Ukrainian                   = 26
    """Ukrainian."""
    SpanishLatinAmerican        = 27
    """Latin American Spanish."""
    Vietnamese                  = 28
    """Vietnamese."""
    # SteamChinaSimplifiedChinese = 29  # not including until this appears on Steamworks

    @classproperty
    def NATIVE_NAME_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's native name."""
        return {
            cls.Arabic:               "العربية",
            cls.Bulgarian:            "езбългарски езикик",
            cls.SimplifiedChinese:    "简体中文",
            cls.TraditionalChinese:   "繁體中文",
            cls.Czech:                "čeština",
            cls.Danish:               "Dansk",
            cls.Dutch:                "Nederlands",
            cls.English:              "English",
            cls.Finnish:              "Suomi",
            cls.French:               "Français",
            cls.German:               "Deutsch",
            cls.Greek:                "Ελληνικά",
            cls.Hungarian:            "Magyar",
            cls.Italian:              "Italiano",
            cls.Japanese:             "日本語",
            cls.Korean:               "한국어",
            cls.Norwegian:            "Norsk",
            cls.Polish:               "Polski",
            cls.Portuguese:           "Português",
            cls.PortugueseBrazil:     "Português-Brasil",
            cls.Romanian:             "Română",
            cls.Russian:              "Русский",
            cls.Spanish:              "Español-España",
            cls.SpanishLatinAmerican: "Español-Latinoamérica",
            cls.Swedish:              "Svenska",
            cls.Thai:                 "ไทย",
            cls.Turkish:              "Türkçe",
            cls.Ukrainian:            "Українська",
            cls.Vietnamese:           "Tiếng Việt",
        }

    @classproperty
    def API_LANGUAGE_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's name from the Steamworks API."""
        return {
            cls.Arabic:               "arabic",
            cls.Bulgarian:            "bulgarian",
            cls.SimplifiedChinese:    "schinese",
            cls.TraditionalChinese:   "tchinese",
            cls.Czech:                "czech",
            cls.Danish:               "danish",
            cls.Dutch:                "dutch",
            cls.English:              "english",
            cls.Finnish:              "finnish",
            cls.French:               "french",
            cls.German:               "german",
            cls.Greek:                "greek",
            cls.Hungarian:            "hungarian",
            cls.Italian:              "italian",
            cls.Japanese:             "japanese",
            cls.Korean:               "koreana",
            cls.Norwegian:            "norwegian",
            cls.Polish:               "polish",
            cls.Portuguese:           "portuguese",
            cls.PortugueseBrazil:     "brazilian",
            cls.Romanian:             "romanian",
            cls.Russian:              "russian",
            cls.Spanish:              "spanish",
            cls.SpanishLatinAmerican: "latam",
            cls.Swedish:              "swedish",
            cls.Thai:                 "thai",
            cls.Turkish:              "turkish",
            cls.Ukrainian:            "ukrainian",
            cls.Vietnamese:           "vietnamese",
        }

    @classproperty
    def WEB_API_MAP(cls: type[Language]) -> Mapping[Language, str]:  # type: ignore
        """A mapping of every language's name from the Web API."""
        return {
            cls.Arabic:               "ar",
            cls.Bulgarian:            "bg",
            cls.SimplifiedChinese:    "zh-CN",
            cls.TraditionalChinese:   "zh-TW",
            cls.Czech:                "cs",
            cls.Danish:               "da",
            cls.Dutch:                "nl",
            cls.English:              "en",
            cls.Finnish:              "fi",
            cls.French:               "fr",
            cls.German:               "de",
            cls.Greek:                "el",
            cls.Hungarian:            "hu",
            cls.Italian:              "it",
            cls.Japanese:             "ja",
            cls.Korean:               "ko",
            cls.Norwegian:            "no",
            cls.Polish:               "pl",
            cls.Portuguese:           "pt",
            cls.PortugueseBrazil:     "pt-BR",
            cls.Romanian:             "ro",
            cls.Russian:              "ru",
            cls.Spanish:              "es",
            cls.SpanishLatinAmerican: "es-419",
            cls.Swedish:              "sv",
            cls.Thai:                 "th",
            cls.Turkish:              "tr",
            cls.Ukrainian:            "uk",
            cls.Vietnamese:           "vn",
        }

    @property
    def native_name(self) -> str:
        """This language's native name."""
        return self.NATIVE_NAME_MAP[self]

    @property
    def api_name(self) -> str:
        """This language's Steamworks name."""
        return self.API_LANGUAGE_MAP[self]

    @property
    def web_api_name(self) -> str:
        """This language's Web API name."""
        return self.WEB_API_MAP[self]

    @classmethod
    def from_str(cls, string: str, /) -> Language:
        try:
            return _REVERSE_API_LANGUAGE_MAP[string.lower()]
        except KeyError:
            return cls._new_member(name=string.title(), value=hash(string.title()))

    @classmethod
    def from_web_api_str(cls, string: str, /) -> Language:
        try:
            return _REVERSE_WEB_API_MAP[string]
        except KeyError:
            return cls._new_member(name=string, value=hash(string))


_REVERSE_API_LANGUAGE_MAP: Final = cast(
    Mapping[str, Language], {value: key for key, value in Language.API_LANGUAGE_MAP.items()}
)
_REVERSE_WEB_API_MAP: Final = cast(
    Mapping[str, Language], {value: key for key, value in Language.WEB_API_MAP.items()}
)


class Currency(IntEnum):
    """All currencies currently supported by Steam."""
    USD = 1
    """The United States Dollar."""
    GBP = 2
    """The British Pound."""
    EUR = 3
    """The Euro."""
    CHF = 4
    """The Swiss Franc."""
    RUB = 5
    """The Russian Ruble."""
    PLN = 6
    """The Polish Zloty."""
    BRL = 7
    """The Brazilian Real."""
    JPY = 8
    """The Japanese Yen."""
    NOK = 9
    """The Norwegian Krone."""
    IDR = 10
    """The Indonesian Rupiah."""
    MYR = 11
    """The Malaysian Ringgit."""
    PHP = 12
    """The Philippine Peso."""
    SGD = 13
    """The Singapore Dollar."""
    THB = 14
    """The Thai Baht."""
    VND = 15
    """The Vietnamese Dong."""
    KRW = 16
    """The South Korean Won."""
    TRY = 17
    """The Turkish Lira."""
    UAH = 18
    """The Ukrainian Hryvnia."""
    MXN = 19
    """The Mexican Peso."""
    CAD = 20
    """The Canadian Dollar."""
    AUD = 21
    """The Australian Dollar."""
    NZD = 22
    """The New Zealand Dollar."""
    CNY = 23
    """The Chinese Yuan."""
    INR = 24
    """The Indian Rupee."""
    CLP = 25
    """The Chilean Peso."""
    PEN = 26
    """The Peruvian Sol."""
    COP = 27
    """The Columbian Peso."""
    ZAR = 28
    """The South African Rand."""
    HKD = 29
    """The Hong Kong Dollar."""
    TWD = 30
    """The Taiwanese New Taiwan Dollar."""
    SAR = 31
    """The Saudi Arabian Saudi Riyal."""
    AED = 32
    """The United Arab Emirates Dirham."""
    SEK = 33
    """The Swedish Krona."""
    ARS = 34
    """The Argentinian Argentine Peso."""
    ILS = 35
    """The Israeli New Shekel."""
    BYN = 36
    """The Belarusian Ruble."""
    KZT = 37
    """The Kazakhstani Tenge."""
    KWD = 38
    """The Kuwaiti Dinar."""
    QAR = 39
    """The Qatari Riyal."""
    CRC = 40
    """The Costa Rican Colón."""
    UYU = 41
    """The Uruguayan Peso."""
    BGN = 42
    """The Bulgarian Lev."""
    HRK = 43
    """The Croatian Kuna."""
    CZK = 44
    """The Czech Koruna."""
    DKK = 45
    """The Danish Krone."""
    HUF = 46
    """The Hungarian Forint."""
    RON = 47
    """The Romanian Leu."""

    @classmethod
    def try_name(cls, name: str, /) -> Self:
        try:
            return cls._member_map_[name]
        except (KeyError, TypeError):
            return cls._new_member(name=name, value=hash(name))


class Realm(IntEnum):
    Unknown = 0
    """Unknown."""
    Global  = 1
    """Global realm."""
    China   = 2
    """China realm."""


class PurchaseResult(IntEnum):
    NoDetail = 0
    AVSFailure = 1
    InsufficientFunds = 2
    ContactSupport = 3
    Timeout = 4
    InvalidPackage = 5
    InvalidPaymentMethod = 6
    InvalidData = 7
    OthersInProgress = 8
    AlreadyPurchased = 9
    WrongPrice = 10
    FraudCheckFailed = 11
    CancelledByUser = 12
    RestrictedCountry = 13
    BadActivationCode = 14
    DuplicateActivationCode = 15
    UseOtherPaymentMethod = 16
    UseOtherFunctionSource = 17
    InvalidShippingAddress = 18
    RegionNotSupported = 19
    AcctIsBlocked = 20
    AcctNotVerified = 21
    InvalidAccount = 22
    StoreBillingCountryMismatch = 23
    DoesNotOwnRequiredApp = 24
    CanceledByNewTransaction = 25
    ForceCanceledPending = 26
    FailCurrencyTransProvider = 27
    FailedCyberCafe = 28
    NeedsPreApproval = 29
    PreApprovalDenied = 30
    WalletCurrencyMismatch = 31
    EmailNotValidated = 32
    ExpiredCard = 33
    TransactionExpired = 34
    WouldExceedMaxWallet = 35
    MustLoginPS3AppForPurchase = 36
    CannotShipToPOBox = 37
    InsufficientInventory = 38
    CannotGiftShippedGoods = 39
    CannotShipInternationally = 40
    BillingAgreementCancelled = 41
    InvalidCoupon = 42
    ExpiredCoupon = 43
    AccountLocked = 44
    OtherAbortableInProgress = 45
    ExceededSteamLimit = 46
    OverlappingPackagesInCart = 47
    NoWallet = 48
    NoCachedPaymentMethod = 49
    CannotRedeemCodeFromClient = 50
    PurchaseAmountNoSupportedByProvider = 51
    OverlappingPackagesInPendingTransaction = 52
    RateLimited = 53
    OwnsExcludedApp = 54
    CreditCardBinMismatchesType = 55
    CartValueTooHigh = 56
    BillingAgreementAlreadyExists = 57
    POSACodeNotActivated = 58
    CannotShipToCountry = 59
    HungTransactionCancelled = 60
    PaypalInternalError = 61
    UnknownGlobalCollectError = 62
    InvalidTaxAddress = 63
    PhysicalProductLimitExceeded = 64
    PurchaseCannotBeReplayed = 65
    DelayedCompletion = 66
    BundleTypeCannotBeGifted = 67
    BlockedByUSGov = 68
    ItemsReservedForCommercialUse = 69
    GiftAlreadyOwned = 70
    GiftInvalidForRecipientRegion = 71
    GiftPricingImbalance = 72
    GiftRecipientNotSpecified = 73
    ItemsNotAllowedForCommercialUse = 74
    BusinessStoreCountryCodeMismatch = 75
    UserAssociatedWithManyCafes = 76
    UserNotAssociatedWithCafe = 77
    AddressInvalid = 78
    CreditCardNumberInvalid = 79
    CannotShipToMilitaryPostOffice = 80
    BillingNameInvalidResemblesCreditCard = 81
    PaymentMethodTemporarilyUnavailable = 82
    PaymentMethodNotSupportedForProduct = 83


class Universe(IntEnum):
    """Steam universes. Each universe is a self-contained Steam instance."""
    Invalid  = 0
    """Invalid."""
    Public   = 1
    """The standard public universe."""
    Beta     = 2
    """Beta universe used inside Valve."""
    Internal = 3
    """Internal universe used inside Valve."""
    Dev      = 4
    """Dev universe used inside Valve."""


class Type(IntEnum):
    """Steam account types."""
    Invalid        = 0
    """Used for invalid Steam IDs."""
    Individual     = 1
    """Single user account."""
    Multiseat      = 2
    """Multiseat (e.g. cybercafe) account."""
    GameServer     = 3
    """Game server account."""
    AnonGameServer = 4
    """Anonymous game server account."""
    Pending        = 5
    """Pending."""
    ContentServer  = 6
    """Valve internal content server account."""
    Clan           = 7
    """Steam clan."""
    Chat           = 8
    """Steam group chat or lobby."""
    ConsoleUser    = 9
    """Fake SteamID for local PSN account on PS3 or Live account on 360, etc."""
    AnonUser       = 10
    """Anonymous user account. (Used to create an account or reset a password)"""


class TypeChar(IntEnum):
    """Steam account type characters."""
    I = Type.Invalid
    """The character used for :class:`~steam.Type.Invalid`."""
    U = Type.Individual
    """The character used for :class:`~steam.Type.Individual`."""
    M = Type.Multiseat
    """The character used for :class:`~steam.Type.Multiseat`."""
    G = Type.GameServer
    """The character used for :class:`~steam.Type.GameServer`."""
    A = Type.AnonGameServer
    """The character used for :class:`~steam.Type.AnonGameServer`."""
    P = Type.Pending
    """The character used for :class:`~steam.Type.Pending`."""
    C = Type.ContentServer
    """The character used for :class:`~steam.Type.ContentServer`."""
    g = Type.Clan
    """The character used for :class:`~steam.Type.Clan`."""
    T = Type.Chat
    """The character used for :class:`~steam.Type.Chat`."""
    L = Type.Chat
    """The character used for :class:`~steam.Type.Chat` (Chat lobby)."""
    c = Type.Chat
    """The character used for :class:`~steam.Type.Clan` (Chat group)."""
    a = Type.AnonUser
    """The character used for :class:`~steam.Type.AnonUser`."""


class Instance(Flags):
    """Steam account instance flags."""
    All          = 0
    """The Instance for all Steam IDs"""
    Desktop      = 1 << 0
    """The Instance for desktop Steam IDs"""
    Console      = 1 << 1
    """The Instance for console  Steam IDs"""
    Web          = 1 << 2
    """The Instance for web Steam IDs"""
    # I have observed these flags for game servers, but I don't know what they are
    Unknown1     = 1 << 3
    Unknown2     = 1 << 4
    Unknown3     = 1 << 5
    Unknown4     = 1 << 6
    Unknown5     = 1 << 7
    Unknown6     = 1 << 8
    Unknown7     = 1 << 9
    Unknown8     = 1 << 10
    Unknown9     = 1 << 11
    Unknown10    = 1 << 12
    Unknown11    = 1 << 13
    Unknown12    = 1 << 14
    # Unknown13    = 1 << 15
    # Unknown14    = 1 << 16
    # 20474 ->   | 2 |   | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # 20475 -> 1 | 2 |   | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # 20476 ->   |   | 4 | 8 | 16 | 32 | 64 | 128 | 256 | 512 | 1024 | 2048 | 16384
    # Type.Chat exclusive flags
    ChatMMSLobby = 1 << 17
    """The Steam ID is for an MMS Lobby."""
    ChatLobby    = 1 << 18
    """The Steam ID is for a Lobby."""
    ChatClan     = 1 << 19
    """The Steam ID is for a Clan."""


class FriendRelationship(IntEnum):
    """The relationship between a user and another user."""
    NONE             = 0
    """The user has no relationship to you."""
    Blocked          = 1
    """The user has been blocked."""
    RequestRecipient = 2
    """The user has requested to be friends with you."""
    Friend           = 3
    """The user is friends with you."""
    RequestInitiator = 4
    """You have requested to be friends with the user."""
    Ignored          = 5
    """You have explicitly blocked this other user from comments/chat/etc."""
    IgnoredFriend    = 6
    """The user has ignored the current user."""


class PersonaState(IntEnum):
    """The status of a user."""
    Offline        = 0
    """The user is not currently logged on."""
    Online         = 1
    """The user is logged on."""
    Busy           = 2
    """The user is on, but busy."""
    Away           = 3
    """The user has been marked as AFK for a short period of time."""
    Snooze         = 4
    """The user has been marked as AFK for a long period of time."""
    LookingToTrade = 5
    """The user is online and wanting to trade."""
    LookingToPlay  = 6
    """The user is online and wanting to play."""
    Invisible      = 7
    """The user is invisible."""


class PersonaStateFlag(Flags):
    """The persona state flags for a user."""
    NONE                 = 0
    """The user has no persona state flags."""
    HasRichPresence      = 1 << 0
    """The user has a rich presence string."""
    InJoinableGame       = 1 << 1
    """The user is in a game and the game has a joinable lobby."""
    Golden               = 1 << 2
    """The user has a golden username and "avatar frame" (this was before avatar frames)."""
    RemotePlayTogether   = 1 << 3
    """The user is playing a game that supports remote play together."""
    ClientTypeWeb        = 1 << 4
    """The user is using the Steam Web App."""
    ClientTypeMobile     = 1 << 8
    """The user is using the Steam Mobile App."""
    ClientTypeTenfoot    = 1 << 10
    """The user is using Steam Big Picture."""
    ClientTypeVR         = 1 << 11
    """The user is using the Steam VR App."""
    LaunchTypeGamepad    = 1 << 12
    """The user is using a gamepad."""
    LaunchTypeCompatTool = 1 << 13
    """The user is using a compatibility tool."""


class CommunityVisibilityState(IntEnum):
    """The visibility of a user's community profile."""
    NONE        = 0
    """The user has no community state."""
    Private     = 1
    """The user has a private profile."""
    FriendsOnly = 2
    """The user has a friends only profile."""
    Public      = 3
    """The user has a public profile."""


class TradeOfferState(IntEnum):
    """The state of a trade offer."""
    Invalid                   = 1
    """The trade offer's state is invalid."""
    Active                    = 2
    """The trade offer is active."""
    Accepted                  = 3
    """The trade offer has been accepted."""
    Countered                 = 4
    """The trade offer has been countered."""
    Expired                   = 5
    """The trade offer has expired."""
    Canceled                  = 6
    """The trade offer has been cancelled."""
    Declined                  = 7
    """The trade offer has be declined by the partner."""
    InvalidItems              = 8
    """The trade offer has invalid items and has been cancelled."""
    ConfirmationNeed          = 9
    """The trade offer needs confirmation."""
    CanceledBySecondaryFactor = 10
    """The trade offer was cancelled by second factor."""
    StateInEscrow             = 11
    """The trade offer is in escrow (hold)."""


class ChatMemberRank(IntEnum):
    """The rank of a chat member."""
    Default = 0
    """Default rank for a chat member."""
    Viewer = 10
    """Viewer rank for a chat member."""
    Guest = 15
    """Guest rank for a chat member."""
    Member = 20
    """Member rank for a chat member."""
    Moderator = 30
    """Moderator rank for a chat member."""
    Officer = 40
    """Officer rank for a chat member."""
    Owner = 50
    """Owner rank for a chat member."""

class ChatMemberJoinState(IntEnum):
    """The join state of a chat member."""
    Default = 0
    NONE = 1
    Joined = 2


class ChatEntryType(IntEnum):
    """The type of chat entry."""
    Invalid          = 0
    """An Invalid Chat entry."""
    Text             = 1
    """A Normal text message from another user."""
    Typing           = 2
    """Another user is typing (not used in multi-user chat)."""
    InviteGame       = 3
    """An Invite from other user into that users current game."""
    LeftConversation = 6
    """A user has left the conversation."""
    Entered          = 7
    """A user has entered the conversation (used in multi-user chat and group chat)."""
    WasKicked        = 8
    """A user was kicked."""
    WasBanned        = 9
    """A user was banned."""
    Disconnected     = 10
    """A user disconnected."""
    HistoricalChat   = 11
    """A chat message from user's chat history or offline message."""
    LinkBlocked      = 14
    """A link was removed by the chat filter."""


class ClanAccountFlags(Flags):
    """Represents """
    Public   = 1 << 0
    """A public clan."""
    Large    = 1 << 1
    """A large clan."""
    Locked   = 1 << 2
    """The clan is locked."""
    Disabled = 1 << 3
    """The clan is disabled."""
    OGG      = 1 << 4
    """An "Official Game Group" (an app clan)."""


class UIMode(IntEnum):
    """The UI mode for a client."""
    Desktop    = 0
    """The UI mode for the desktop client."""
    BigPicture = 1
    """The UI mode for big picture mode."""
    Mobile     = 2
    """The UI mode for mobile."""
    Web        = 3
    """The UI mode for the web client."""


class ReviewType(IntEnum):
    """The type of review."""
    NONE                   = 0
    """No reviews."""
    OverwhelminglyNegative = 1
    """0 - 19% positive reviews and few of them."""
    VeryNegative           = 2
    """0 - 19% positive reviews."""
    Negative               = 3
    """0 - 39% positive reviews."""
    MostlyNegative         = 4
    """20 - 39% positive reviews but few of them."""
    Mixed                  = 5
    """40 - 69% positive reviews."""
    MostlyPositive         = 6
    """70 - 79% positive reviews."""
    Positive               = 7
    """80 - 100% positive reviews but few of them."""
    VeryPositive           = 8
    """94 - 80% positive reviews."""
    OverwhelminglyPositive = 9
    """95 - 100% positive reviews."""


class GameServerRegion(IntEnum):
    """The region of a game server."""
    NONE         = -1
    """No set game region."""
    USEastCoast  = 0
    """A server on the USA's East Coast."""
    USWestCoast  = 1
    """A server on the USA's West Coast."""
    SouthAmerica = 2
    """A server in South America."""
    Europe       = 3
    """A server in Europe."""
    Asia         = 4
    """A server in Asia."""
    Australia    = 5
    """A server in Australia."""
    MiddleEast   = 6
    """A server in the Middle East."""
    Africa       = 7
    """A server in Africa."""
    World        = 255
    """A server somewhere in the world."""


class EventType(IntEnum):
    """The type of an event."""
    Other                  = 1
    """An unspecified event."""
    Game                   = 2
    """A game event."""
    Party                  = 3
    """A party event."""
    Meeting                = 4
    """An important meeting."""
    SpecialCause           = 5
    """An event for a special cause."""
    MusicAndArts           = 6
    """A music or art event."""
    Sports                 = 7
    """A sporting event."""
    Trip                   = 8
    """A clan trip."""
    Chat                   = 9
    """A chat event."""
    GameRelease            = 10
    """A game release event."""
    Broadcast              = 11
    """A broadcast event."""
    SmallUpdate            = 12
    """A small update event."""
    PreAnnounceMajorUpdate = 13
    """A pre-announcement for a major update event."""
    MajorUpdate            = 14
    """A major update event."""
    DLCRelease             = 15
    """A dlc release event."""
    FutureRelease          = 16
    """A future release event."""
    ESportTournamentStream = 17
    """An e-sport tournament stream event."""
    DevStream              = 18
    """A developer stream event."""
    FamousStream           = 19
    """A famous stream event."""
    GameSales              = 20
    """A game sales event."""
    GameItemSales          = 21
    """A game item sales event."""
    InGameBonusXP          = 22
    """An in game bonus xp event."""
    InGameLoot             = 23
    """An in game loot event."""
    InGamePerks            = 24
    """An in game perks event."""
    InGameChallenge        = 25
    """An in game challenge event."""
    InGameContest          = 26
    """An in game contest event."""
    IRL                    = 27
    """An in real life event."""
    News                   = 28
    """A news event."""
    BetaRelease            = 29
    """A beta release event."""
    InGameContentRelease   = 30
    """An in game content release event."""
    FreeTrial              = 31
    """A free trial event."""
    SeasonRelease          = 32
    """A season release event."""
    SeasonUpdate           = 33
    """A season update event."""
    Crosspost              = 34
    """A cross post event."""
    InGameGeneral          = 35
    """An in game general event."""


class ProfileItemType(IntEnum):
    """The type of profile item."""
    Invalid                   = 0
    """An invalid item type."""
    RareAchievementShowcase   = 1
    """A rare achievements showcase."""
    GameCollector             = 2
    """A game collector section."""
    ItemShowcase              = 3
    """An item showcase."""
    TradeShowcase             = 4
    """A trade info showcase."""
    Badges                    = 5
    """A badges showcase."""
    FavouriteGame             = 6
    """A favourite game section."""
    ScreenshotShowcase        = 7
    """A screenshot showcase."""
    CustomText                = 8
    """A custom text section."""
    FavouriteGroup            = 9
    """A favourite game showcase."""
    Recommendation            = 10
    """A review showcase."""
    WorkshopItem              = 11
    """A workshop item showcase."""
    MyWorkshop                = 12
    """A showcase of a workshop item made by profile's owner."""
    ArtworkShowcase           = 13
    """An artwork showcase."""
    VideoShowcase             = 14
    """A video showcase."""
    Guides                    = 15
    """A guide showcase."""
    MyGuides                  = 16
    """A showcase of the profile's owner's guides."""
    Achievements              = 17
    """The owner's profile's achievements."""
    Greenlight                = 18
    """A greenlight showcase."""
    MyGreenlight              = 19
    """A showcase of a greenlight game the profiles' owner has made."""
    Salien                    = 20
    """A salien showcase."""
    LoyaltyRewardReactions    = 21
    """A loyalty reward showcase."""
    SingleArtworkShowcase     = 22
    """A single artwork showcase."""
    AchievementsCompletionist = 23
    """An achievements completeionist showcase."""
    Replay = 24
    """A Steam Replay showcase."""


class ProfileCustomisationStyle(IntEnum):
    Default      = 0
    Selected     = 1
    Rarest       = 2
    MostRecent   = 3
    Random       = 4
    HighestRated = 5


class CommunityItemClass(IntEnum):
    """An item class."""
    Invalid               = 0
    """An invalid item class."""
    Badge                 = 1
    """A badge."""
    GameCard              = 2
    """A game card."""
    ProfileBackground     = 3
    """A profile background."""
    Emoticon              = 4
    """An emoticon."""
    BoosterPack           = 5
    """A booster pack."""
    Consumable            = 6
    """A consumable."""
    GameGoo               = 7
    """Game goo."""
    ProfileModifier       = 8
    """A profile modifier."""
    Scene                 = 9
    """A scene."""
    SalienItem            = 10
    """A salien item."""
    Sticker               = 11
    """A sticker."""
    ChatEffect            = 12
    """A chat effect."""
    MiniProfileBackground = 13
    """A mini profile background."""
    AvatarFrame           = 14
    """An avatar frame."""
    AnimatedAvatar        = 15
    """An animated avatar."""
    SteamDeckKeyboardSkin = 16
    """A Steam Deck keyboard skin."""


class ProfileItemEquippedFlag(Flags):
    """Flags for a profile item."""
    FullScreen = 1
    """The item is equipped in full screen mode."""


class DepotFileFlag(Flags):
    """Flags for a depot file."""
    File                = 0
    """A file."""
    UserConfig          = 1 << 0
    """A user configuration file."""
    VersionedUserConfig = 1 << 1
    """A versioned user configuration file."""
    Encrypted           = 1 << 2
    """An encrypted file."""
    ReadOnly            = 1 << 3
    """A read only file."""
    Hidden              = 1 << 4
    """A hidden file."""
    Executable          = 1 << 5
    """An executable file."""
    Directory           = 1 << 6
    """A directory."""
    CustomExecutable    = 1 << 7
    """A custom executable file."""
    InstallScript       = 1 << 8
    """An install script file."""
    Symlink             = 1 << 9
    """A symlink file."""


TYPE_TRANSFORM_MAP: Final = cast(Mapping[str, str], {
    "Dlc": "DLC",
})


class AppType(Flags):
    """App type."""
    Game        = 1 << 0
    """A playable game, default type."""
    Application = 1 << 1
    """A software application."""
    Tool        = 1 << 2
    """An SDK, editor or dedicated server."""
    Demo        = 1 << 3
    """A game demo."""
    DLC         = 1 << 5
    """A piece of downloadable content."""
    Guide       = 1 << 6
    """A game guide or PDF etc."""
    Driver      = 1 << 7
    """A hardware driver updater (ATI, Razor etc.)"""
    Config      = 1 << 8
    """A hidden app used to config Steam features (backpack, sales, etc.)."""
    Hardware    = 1 << 9
    """A hardware device (Steam Machine, Steam Controller, Steam Link, etc.)."""
    Franchise   = 1 << 10
    """A hub for collections of multiple apps, e.g. films, series, games.."""
    Video       = 1 << 11
    """A video component of either a Film or TVSeries (may be the feature, an episode, preview, making-of, etc.)"""
    Plugin      = 1 << 12
    """A plug-in type for another App."""
    Music       = 1 << 13
    """A piece of music."""
    Series      = 1 << 14
    """A container app for video series."""
    Comic       = 1 << 15
    """A comic."""
    Beta        = 1 << 16
    """A beta for a game."""
    Media = 1 << 17
    """Legacy Media"""

    Shortcut    = 1 << 30
    """A shortcut to another app, client side only."""
    DepotOnly   = -1 << 31
    """A placeholder since depots and apps share the same namespace."""

    @classmethod
    def from_str(cls, name: str, /) -> AppType:
        types = iter(name.split(","))
        type = next(types).strip().title()
        self = cls[TYPE_TRANSFORM_MAP.get(type, type)]
        for type in types:
            type = type.strip().title()
            self |= cls[TYPE_TRANSFORM_MAP.get(type, type)]
        return self


class LicenseFlag(Flags):
    """Flags for a license."""
    NONE                         = 0
    """No flags."""
    Renew                        = 1 << 0
    """License needs renewing."""
    RenewalFailed                = 1 << 1
    """License renewal failed."""
    Pending                      = 1 << 2
    """Owns license, but transaction is still pending. Can't install or play yet."""
    Expired                      = 1 << 3
    """License is expired."""
    CancelledByUser              = 1 << 4
    """License is cancelled by user."""
    CancelledByAdmin             = 1 << 5
    """License is cancelled by an admin."""
    LowViolenceContent           = 1 << 6
    """License is only for a low violence version."""
    ImportedFromSteam2           = 1 << 7
    """License was imported from Steam 2."""
    ForceRunRestriction          = 1 << 8
    """License has a force run restriction."""
    RegionRestrictionExpired     = 1 << 9
    """License has a region restriction that has expired."""
    CancelledByFriendlyFraudLock = 1 << 10
    """License is cancelled by a friendly fraud lock."""
    NotActivated                 = 1 << 11
    """License is not activated."""


class LicenseType(IntEnum):
    """Types of licenses."""
    NoLicense                             = 0
    """No license."""
    SinglePurchase                        = 1
    """A single purchase license."""
    SinglePurchaseLimitedUse              = 2
    """A single purchase limited use license."""
    RecurringCharge                       = 3
    """A subscription based license."""
    RecurringChargeLimitedUse             = 4
    """A subscription based limited use license."""
    RecurringChargeLimitedUseWithOverAges = 5
    """A subscription based limited use license with over ages."""
    RecurringOption                       = 6
    """A subscription based option license."""
    LimitedUseDelayedActivation           = 7
    """A limited use delayed activation license."""


class BillingType(IntEnum):
    """Types of billing ."""
    NoCost                 = 0
    """No cost."""
    BillOnceOnly           = 1
    """Bill once only."""
    BillMonthly            = 2
    """Bill monthly."""
    ProofOfPrepurchaseOnly = 3
    """Proof of pre-purchase only."""
    GuestPass              = 4
    """Guest pass."""
    HardwarePromo          = 5
    """Hardware promotion."""
    Gift                   = 6
    """Gift."""
    AutoGrant              = 7
    """Automatic grant."""
    OEMTicket              = 8
    """Original Equipment Manufacturer ticket."""
    RecurringOption        = 9
    """Recurring option."""
    BillOnceOrCDKey        = 10
    """Bill once or CD key."""
    Repurchasable          = 11
    """Re-purchasable."""
    FreeOnDemand           = 12
    """Free on demand."""
    Rental                 = 13
    """Rental."""
    CommercialLicense      = 14
    """Commercial license."""
    FreeCommercialLicense  = 15
    """Free commercial license."""
    NumBillingTypes        = 16
    """Number of billing types."""


class PaymentMethod(IntEnum):
    """The payment method used for a purchase."""
    NONE                   = 0
    """No payment method."""
    ActivationCode         = 1
    """Activation code."""
    CreditCard             = 2
    """Credit card."""
    Giropay                = 3
    """Giropay."""
    PayPal                 = 4
    """PayPal."""
    Ideal                  = 5
    """iDEAL."""
    PaySafeCard            = 6
    """PaySafeCard."""
    Sofort                 = 7
    """Sofort."""
    GuestPass              = 8
    """Guest pass."""
    WebMoney               = 9
    """WebMoney."""
    MoneyBookers           = 10
    """MoneyBookers."""
    AliPay                 = 11
    """AliPay."""
    Yandex                 = 12
    """Yandex."""
    Kiosk                  = 13
    """Kiosk."""
    Qiwi                   = 14
    """Qiwi."""
    GameStop               = 15
    """GameStop."""
    HardwarePromo          = 16
    """Hardware promotion."""
    MoPay                  = 17
    """MoPay."""
    BoletoBancario         = 18
    """Boleto Bancario."""
    BoaCompraGold          = 19
    """BoaCompra Gold."""
    BancoDoBrasilOnline    = 20
    """Banco do Brasil Online."""
    ItauOnline             = 21
    """Itau Online."""
    BradescoOnline         = 22
    """Bradesco Online."""
    Pagseguro              = 23
    """Pagseguro."""
    VisaBrazil             = 24
    """Visa Brazil."""
    AmexBrazil             = 25
    """American Express Brazil."""
    Aura                   = 26
    """Aura."""
    Hipercard              = 27
    """Hipercard."""
    MastercardBrazil       = 28
    """Mastercard Brazil."""
    DinersCardBrazil       = 29
    """Diners Card Brazil."""
    AuthorizedDevice       = 30
    """Authorized device."""
    MOLPoints              = 31
    """MOLPoints."""
    ClickAndBuy            = 32
    """Click and Buy."""
    Beeline                = 33
    """Beeline."""
    Konbini                = 34
    """Konbini."""
    EClubPoints            = 35
    """E-Club Points."""
    CreditCardJapan        = 36
    """Credit card Japan."""
    BankTransferJapan      = 37
    """Bank transfer Japan."""
    PayEasy                = 38
    """PayEasy."""
    Zong                   = 39
    """Zong."""
    CultureVoucher         = 40
    """Culture voucher."""
    BookVoucher            = 41
    """Book voucher."""
    HappymoneyVoucher      = 42
    """Happy money voucher."""
    ConvenientStoreVoucher = 43
    """Convenient store voucher."""
    GameVoucher            = 44
    """Game voucher."""
    Multibanco             = 45
    """Multibanco."""
    Payshop                = 46
    """Payshop."""
    MaestroBoaCompra       = 47
    """Maestro BoaCompra."""
    OXXO                   = 48
    """OXXO."""
    ToditoCash             = 49
    """Todito Cash."""
    Carnet                 = 50
    """Carnet."""
    SPEI                   = 51
    """SPEI."""
    ThreePay               = 52
    """Three Pay."""
    IsBank                 = 53
    """Is Bank."""
    Garanti                = 54
    """Garanti."""
    Akbank                 = 55
    """Akbank."""
    YapiKredi              = 56
    """Yapi"""
    Halkbank               = 57
    """Halkbank."""
    BankAsya               = 58
    """Bank Asya."""
    Finansbank             = 59
    """Finansbank."""
    DenizBank              = 60
    """DenizBank."""
    PTT                    = 61
    """PTT."""
    CashU                  = 62
    """CashU."""
    AutoGrant              = 64
    """Automatic grant."""
    WebMoneyJapan          = 65
    """WebMoney Japan."""
    OneCard                = 66
    """OneCard."""
    PSE                    = 67
    """PSE."""
    Exito                  = 68
    """Exito."""
    Efecty                 = 69
    """Efecty."""
    Paloto                 = 70
    """Paloto."""
    PinValidda             = 71
    """Pin Validda."""
    MangirKart             = 72
    """Mangir Kart."""
    BancoCreditoDePeru     = 73
    """Banco Credito de Peru."""
    BBVAContinental        = 74
    """BBVA Continental."""
    SafetyPay              = 75
    """SafetyPay."""
    PagoEfectivo           = 76
    """Pago Efectivo."""
    Trustly                = 77
    """Trustly."""
    UnionPay               = 78
    """UnionPay."""
    BitCoin                = 79
    """BitCoin."""
    Wallet                 = 128
    """Wallet."""
    Valve                  = 129
    """Valve."""
    MasterComp             = 130
    """Master Comp."""
    Promotional            = 131
    """Promotional."""
    MasterSubscription     = 134
    """Master Subscription."""
    Payco                  = 135
    """Payco."""
    MobileWalletJapan      = 136
    """Mobile Wallet Japan."""
    OEMTicket              = 256
    """Original Equipment Manufacturer Ticket."""
    Split                  = 512
    """Split."""
    Complimentary          = 1024
    """Complimentary."""


class PackageStatus(IntEnum):
    """Package status."""
    Available   = 0
    """Available."""
    Preorder    = 1
    """Pre-order."""
    Unavailable = 2
    """Unavailable."""
    Invalid     = 3
    """Invalid."""


class PublishedFileRevision(IntEnum):
    """Published file revisions."""
    Default               = 0
    """The default revision."""
    Latest                = 1
    """The latest revision."""
    ApprovedSnapshot      = 2
    """The approved snapshot."""
    ApprovedSnapshotChina = 3
    """The approved snapshot for China."""
    RejectedSnapshot      = 4
    """The rejected snapshot."""
    RejectedSnapshotChina = 5
    """The rejected snapshot for China."""


class PublishedFileQueryType(IntEnum):
    """Ways you can query for UGC items (published files)."""
    RankedByVote                                  = 0
    """Ranked by vote."""
    RankedByPublicationDate                       = 1
    """Ranked by publication date."""
    AcceptedForGameRankedByAcceptanceDate         = 2
    """Accepted for game ranked by acceptance date."""
    RankedByTrend                                 = 3
    """Ranked by trend."""
    FavouritedByFriendsRankedByPublicationDate    = 4
    """Favourited by friends ranked by publication date."""
    CreatedByFriendsRankedByPublicationDate       = 5
    """Created by friends ranked by publication date."""
    RankedByNumTimesReported                      = 6
    """Ranked by number of times reported."""
    CreatedByFollowedUsersRankedByPublicationDate = 7
    """Created by followed users ranked by publication date."""
    NotYetRated                                   = 8
    """Not yet rated."""
    RankedByTotalUniqueSubscriptions              = 9
    """Ranked by total unique subscriptions."""
    RankedByTotalVotesAsc                         = 10
    """Ranked by total votes ascending."""
    RankedByVotesUp                               = 11
    """Ranked by votes up."""
    RankedByTextSearch                            = 12
    """Ranked by text search."""
    RankedByPlaytimeTrend                         = 13
    """Ranked by playtime trend."""
    RankedByTotalPlaytime                         = 14
    """Ranked by total playtime."""
    RankedByAveragePlaytimeTrend                  = 15
    """Ranked by average playtime trend."""
    RankedByLifetimeAveragePlaytime               = 16
    """Ranked by lifetime average playtime."""
    RankedByPlaytimeSessionsTrend                 = 17
    """Ranked by playtime sessions trend."""
    RankedByLifetimePlaytimeSessions              = 18
    """Ranked by lifetime playtime sessions."""
    RankedByInappropriateContentRating            = 19
    """Ranked by inappropriate content rating."""
    RankedByBanContentCheck                       = 20
    """Ranked by ban content check."""
    RankedByLastUpdatedDate                       = 21
    """Ranked by last updated date."""


class PublishedFileType(IntEnum):
    """Published file types."""
    Community              = 0
    """Normal Workshop item that can be subscribed to"""
    Microtransaction       = 1
    """Workshop item that is meant to be voted on for the purpose of selling in-game"""
    Collection             = 2
    """A collection of Workshop or Greenlight items"""
    Art                    = 3
    """Artwork"""
    Video                  = 4
    """External video"""
    Screenshot             = 5
    """Screenshot"""
    Game                   = 6
    """Greenlight game entry"""
    Software               = 7
    """Greenlight software entry"""
    Concept                = 8
    """Greenlight concept"""
    WebGuide               = 9
    """Steam web guide"""
    IntegratedGuide        = 10
    """Application integrated guide"""
    Merch                  = 11
    """Workshop merchandise meant to be voted on for the purpose of being sold"""
    ControllerBinding      = 12
    """Steam Controller bindings"""
    SteamworksAccessInvite = 13
    """Internal"""
    SteamVideo             = 14
    """Steam video"""
    GameManagedItem        = 15
    """Managed completely by the game, not the user, and not shown on the web"""


class PublishedFileVisibility(IntEnum):
    """Published file visibility."""
    Public      = 0
    """Public."""
    FriendsOnly = 1
    """Friends only."""
    Private     = 2
    """Private."""
    Unlisted    = 3
    """Unlisted."""


class PublishedFileQueryFileType(IntEnum):
    """The way that a shared file can be queried by QueryFile."""
    Items                   = 0
    """Items."""
    Collections             = 1
    """A collection of Workshop items."""
    Art                     = 2
    """Artwork."""
    Videos                  = 3
    """Videos."""
    Screenshots             = 4
    """Screenshots."""
    CollectionEligible      = 5
    """Items that can be put inside a collection."""
    Games                   = 6
    """Unused."""
    Software                = 7
    """Unused."""
    Concepts                = 8
    """Unused."""
    GreenlightItems         = 9
    """Unused."""
    AllGuides               = 10
    """Guides."""
    WebGuides               = 11
    """Steam web guide."""
    IntegratedGuides        = 12
    """Application integrated guide."""
    UsableInGame            = 13
    """Workshop items that can be used right away by the user."""
    Merch                   = 14
    """Workshop merchandise meant to be voted on for the purpose of being sold."""
    ControllerBindings      = 15
    """Steam Controller bindings."""
    SteamworksAccessInvites = 16
    """Used internally."""
    ForSaleInGame           = 17
    """Workshop items that can be sold in-game."""
    ReadyToUseItems         = 18
    """Workshop items that can be used right away by the user."""
    WorkshopShowcases       = 19
    """Workshop showcases."""
    GameManagedItems        = 20
    """Managed completely by the game, not the user, and not shown on the web."""


class LeaderboardDataRequest(IntEnum):
    """The way that a leaderboard can be queried."""
    Global = 0
    """Global leaderboard."""
    GlobalAroundUser = 1
    """Global leaderboard around user."""
    Friends = 2
    """Friends leaderboard."""
    Users = 3
    """Users leaderboard."""


class LeaderboardSortMethod(IntEnum):
    """The way that a leaderboard can be sorted."""
    NONE = 0
    """Default sort method."""
    Ascending = 1
    """Ascending sort method."""
    Descending = 2
    """Descending sort method."""


class LeaderboardDisplayType(IntEnum):
    """The way that a leaderboard can be displayed."""
    NONE = 0
    """Default display type."""
    Numeric = 1
    """Numeric display type."""
    TimeSeconds = 2
    """Time in seconds display type."""
    TimeMilliseconds = 3
    """Time in milliseconds display type."""


class LeaderboardUploadScoreMethod(IntEnum):
    """The way that a leaderboard can be uploaded to."""
    NONE = 0
    """Default upload method."""
    KeepBest = 1
    """Keep best upload method."""
    ForceUpdate = 2
    """Force update upload method."""


class CommunityDefinitionItemType(IntEnum):
    """The type of a community item."""
    NONE = 0


class AuthSessionResponse(IntEnum):
    """The response from an app ticket authentication."""
    NotSent                      = -1
    """Not sent. Not a valid response from Steam."""
    OK                           = 0
    UserNotConnectedToSteam      = 1
    NoLicenseOrExpired           = 2
    VACBanned                    = 3
    LoggedInElseWhere            = 4
    VACCheckTimedOut             = 5
    AuthTicketCanceled           = 6
    AuthTicketInvalidAlreadyUsed = 7
    AuthTicketInvalid            = 8
    PublisherIssuedBan           = 9


class ContentDescriptor(IntEnum):
    NudityOrSexualContent = 1
    FrequentNudityOrSexualContent = 1
    FrequentViolenceOrGore = 2
    AdultOnlySexualContent = 3
    StrongSexualContent = 3
    GratuitousSexualContent = 4
    AnyMatureContent = 5


class has_associated_flag(int):
    flag: int
    def __new__(cls, value: int, flag: int) -> Self:
        self = super().__new__(cls, value)
        super().__setattr__(self, "flag", flag)
        return self


class UserNewsType(IntEnum):
    FriendAdded                       = has_associated_flag(1,    1 << 0)
    AchievementUnlocked               = has_associated_flag(2,    1 << 1)
    ReceivedNewGame                   = has_associated_flag(3,    1 << 2)
    JoinedGroup                       = has_associated_flag(4,    1 << 3)
    CommentByMe                       = has_associated_flag(5,    0)
    FriendRemoved                     = has_associated_flag(6,    0)
    GroupCreated                      = has_associated_flag(7,    0)
    CommentOnMe                       = has_associated_flag(8,    0)
    AddedGameToWishlist               = has_associated_flag(9,    1 << 7)
    Review                            = has_associated_flag(10,   1 << 8)
    FilePublishedScreenshot           = has_associated_flag(13,   1 << 9)
    FilePublishedVideo                = has_associated_flag(14,   1 << 10)
    FilePublishedWorkshopItem         = has_associated_flag(15,   1 << 11)
    Post                              = has_associated_flag(16,   1 << 12)
    FilePublishedCollection           = has_associated_flag(17,   1 << 13)
    FilePublishedGreenlightGame       = has_associated_flag(18,   0)
    FilePublishedWorkshopAnnouncement = has_associated_flag(19,   0)
    FilePublishedWebGuide             = has_associated_flag(20,   1 << 13)
    FilePublishedScreenshotTagged     = has_associated_flag(21,   0)
    FilePublishedArt                  = has_associated_flag(22,   1 << 13)
    FileFavourited                    = has_associated_flag(23,   1 << 14)
    PlayedGameFirstTime               = has_associated_flag(30,   1 << 2)

    ClanAchievement                   = has_associated_flag(1001, 1 << 16)
    PostedAnnouncement                = has_associated_flag(1002, 1 << 17)
    ScheduledEvent                    = has_associated_flag(1003, 1 << 18)
    SelectedNewPOTW                   = has_associated_flag(1004, 1 << 19)
    PromotedNewAdmin                  = has_associated_flag(1005, 1 << 20)
    MessageOnClanPage                 = has_associated_flag(1006, 1 << 21)
    CuratorReview                     = has_associated_flag(1007, 1 << 22)

    @property
    def flag(self) -> int:
        return self.value.flag

    def __or__(self, other: SupportsInt) -> Self:
        cls = self.__class__
        value = self.flag | other.flag if isinstance(other, cls) else self.flag | int(other)
        return cls._new_member(name=f"{self.name} | {getattr(other, 'name', other)}", value=has_associated_flag(-1, value))

    def __and__(self, other: SupportsInt) -> Self:
        cls = self.__class__
        value = self.flag & other.flag if isinstance(other, cls) else self.flag & int(other)
        return cls._new_member(name=f"{self.name} & {getattr(other, 'name', other)}", value=has_associated_flag(-1, value))

    @classproperty
    def App(cls: type[Self]) -> UserNewsType:  # type: ignore
        """Mimics the Steam Client's feed when an app is selected."""
        return (
            cls.AchievementUnlocked
            | cls.FilePublishedScreenshot
            | cls.FilePublishedVideo
            | cls.Post
            | cls.Review
            | cls.CuratorReview
            | cls.AddedGameToWishlist
            | cls.PlayedGameFirstTime
        )

    @classproperty
    def Friend(cls: type[Self]) -> UserNewsType:  # type: ignore
        """Mimics the Steam Client's feed when in the friend activity tab."""
        return cls.FriendAdded | cls.App

    @classproperty
    def All(cls: type[Self]) -> UserNewsType:  # type: ignore
        """Returns a member with every flag set."""
        return reduce(cls.__or__, (member for member in cls if member.flag != 0))

# fmt: on


if __debug__:
    _ENUM_NAMES = {enum.__name__ for enum in Enum.__subclasses__() + IntEnum.__subclasses__() + Flags.__subclasses__()}
    assert _ENUM_NAMES.issubset(__all__), f"__all__ is not complete, missing {_ENUM_NAMES - set(__all__)}"



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Final, Literal, NamedTuple, Protocol, runtime_checkable

from typing_extensions import TypeVar

from .types.id import AssetID
from .types.user import UserT
from .utils import DateTime, get

if TYPE_CHECKING:
    from .app import STEAM, PartialApp
    from .models import CDNAsset
    from .state import ConnectionState
    from .trade import Item
    from .types import user


__all__ = (
    "AppBadge",
    "BadgeProgress",
    "FavouriteBadge",
    "UserBadge",
    "UserBadges",
)

AppT = TypeVar("AppT", bound="PartialApp", default="PartialApp", covariant=True)


@runtime_checkable
class BaseBadge(Protocol[AppT]):
    __slots__ = ("id", "level", "app", "_state")
    _state: ConnectionState
    id: int
    """The badge's ID."""
    level: float
    """The badge's level. :class:`int` or ``float("inf")``"""
    app: Final[AppT]
    """The app associated with the badge."""

    def __init__(self, state: ConnectionState, id: int, level: float, app: AppT) -> None:
        self._state = state
        self.id = id
        self.level = level
        self.app = app

    def __repr__(self) -> str:
        attrs = ("id", "level", "app")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, BaseBadge) and self.id == other.id and self.app == other.app and self.level == other.level
        )

    def __hash__(self) -> int:
        return hash((self.id, self.app, self.level))

    async def _fetch_app_badge(self) -> AppBadge[AppT]:
        badges = await self.app.badges()
        badge = get(badges, id=self.id)
        assert badge is not None
        return badge

    async def name(self) -> str:
        """Fetches the name of this badge."""
        app_badge = await self._fetch_app_badge()
        return app_badge._name

    async def icon(self) -> CDNAsset:
        """Fetches the URL of this badge."""
        app_badge = await self._fetch_app_badge()
        return app_badge._icon


class AppBadge(BaseBadge[AppT]):
    """Represents a badge for an app."""

    __slots__ = ("_name", "_icon")

    def __init__(self, state: ConnectionState, id: int, name: str, app: AppT, icon: CDNAsset, level: float = 1):
        super().__init__(state, id, level, app)
        self._name = name
        self._icon = icon

    async def name(self) -> str:
        return self._name

    async def icon(self) -> CDNAsset:
        return self._icon

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} name()={self._name!r} level={self.level} app={self.app!r}>"


class BadgeProgress(NamedTuple):
    quest_id: int
    complete: bool


@runtime_checkable
class BaseOwnedBadge(BaseBadge[AppT], Protocol[AppT, UserT]):
    __slots__ = ("owner", "community_item_id")
    owner: Final[UserT]
    """The user who owns this badge."""
    community_item_id: AssetID | None
    r"""The badge's community item ID. By itself this doesn't correspond to anything in the :attr:`owner`\s inventory.
     - ``-1`` is the associated emoticon created by crafting the badge.
     - ``-2`` is the associated background created by crafting the badge.
    """

    def __init__(
        self, state: ConnectionState, id: int, level: int, app: AppT, owner: UserT, community_item_id: int | str | None
    ):
        super().__init__(state, id, level, app)
        self.owner = owner
        self.community_item_id = AssetID(int(community_item_id)) if community_item_id is not None else None

    async def _from_inventory(self, offset: int) -> Item[UserT]:
        from .app import STEAM

        if self.community_item_id is None:
            raise ValueError("This badge doesn't have an associated item.")

        inventory = await self.owner.inventory(STEAM)
        item = get(inventory, id=self.community_item_id - offset)
        if item is None:
            raise ValueError("This badge doesn't have an associated item.")
        return item

    async def emoticon(self) -> Item[UserT]:
        """The emoticon for this badge as an :class:`Item`.

        Raises
        ------
        ValueError
            The emoticon for this badge doesn't exist.
        """
        return await self._from_inventory(1)

    async def background(self) -> Item[UserT]:
        """The background for this badge as an :class:`Item`.

        Raises
        ------
        ValueError
            The background for this badge doesn't exist.
        """
        return await self._from_inventory(2)

    async def progress(self: BaseOwnedBadge[Literal[STEAM]]) -> list[BadgeProgress]:
        """Get the progress

        Returns
        -------
        .. source:: BadgeProgress
        """
        data = await self._state.http.get_user_community_badge_progress(self.owner.id64, self.id)
        return [BadgeProgress(badge["questid"], badge["completed"]) for badge in data]


class FavouriteBadge(BaseOwnedBadge[AppT, UserT]):
    """Represents a User's favourite badge."""

    __slots__ = ("border_colour", "type")

    def __init__(
        self,
        state: ConnectionState,
        id: int,
        level: int,
        app: AppT,
        owner: UserT,
        community_item_id: int,
        type: int,
        border_colour: int,
    ):
        super().__init__(state, id, level, app, owner, community_item_id)
        self.type = type
        """The badge's type."""
        self.border_colour = border_colour
        """The badge's border colour."""


class UserBadge(BaseOwnedBadge["PartialApp | Literal[STEAM]", UserT]):
    """Represents a Steam badge on a user's profile."""

    __slots__ = ("xp", "scarcity", "completed_at")

    def __init__(self, state: ConnectionState, owner: UserT, data: user.UserBadgeBadge):
        from .app import STEAM, PartialApp

        super().__init__(
            state,
            data["badgeid"],
            data["level"],
            PartialApp(state, id=data["appid"]) if "appid" in data else STEAM,
            owner,
            data.get("communityitemid"),
        )
        self.xp = data["xp"]
        """The badge's XP."""
        self.completed_at = DateTime.from_timestamp(data["completion_time"])
        """The time the badge was completed at."""
        self.scarcity = data["scarcity"]
        """The scarcity of the badge. The lower this value, the rarer the badge."""


class UserBadges(Sequence[UserBadge[UserT]]):
    """Represents a Steam :class:`~steam.User`'s badges/level.

    .. container:: operations

        .. describe:: len(x)

            Returns the number of badges.

        .. describe:: iter(x)

            Returns an iterator over the badges.

        .. describe:: y in x

            Returns whether the badge is in the badges.

        .. describe:: x[y]

            Returns the badge at the given index.
    """

    __slots__ = (
        "user",
        "xp",
        "level",
        "badges",
        "xp_needed_to_level_up",
        "xp_needed_for_current_level",
    )

    def __init__(self, state: ConnectionState, user: UserT, data: user.UserBadges):
        self.user = user
        self.level = data["player_level"]
        """The user's level."""
        self.xp = data["player_xp"]
        """The users's XP."""
        self.xp_needed_to_level_up = data["player_xp_needed_to_level_up"]
        """The amount of XP the user needs to level up."""
        self.xp_needed_for_current_level = data["player_xp_needed_current_level"]
        """The amount of XP the user's current level requires to achieve."""
        self.badges: Sequence[UserBadge[UserT]] = [UserBadge(state, user, data) for data in data["badges"]]
        """A list of the user's badges."""

    def __repr__(self) -> str:
        attrs = ("level", "xp")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f'<UserBadges {" ".join(resolved)}>'

    def __len__(self) -> int:
        return len(self.badges)

    if not TYPE_CHECKING:

        def __getitem__(self, idx):
            return self.badges[idx]



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from typing_extensions import TypeVar

from ._const import URL, impl_eq_via_id
from .app import PartialApp
from .clan import PartialClan
from .enums import Language
from .package import PartialPackage
from .types.id import BundleID, Intable

if TYPE_CHECKING:
    from .state import ConnectionState
    from .store import BundleStoreItem
    from .types import bundle

__all__ = (
    "Bundle",
    "PartialBundle",
    "FetchedBundle",
)

NameT = TypeVar("NameT", bound=str | None, default=str | None, covariant=True)


@impl_eq_via_id
class Bundle(Generic[NameT]):
    """Represents a bundle of apps and packages. Read more on :works:`steamworks <store/application/bundles>`"""

    __slots__ = ("id", "name")

    def __init__(self, *, id: Intable, name: NameT = None) -> None:
        self.id = BundleID(int(id))
        """The bundle's ID."""
        self.name = name
        """The bundle's name."""

    def __repr__(self) -> str:
        return f"Bundle(id={self.id}, name={self.name!r})"

    @property
    def url(self) -> str:
        """The bundle's store page URL."""
        return f"{URL.STORE}/bundle/{self.id}"


class PartialBundle(Bundle[NameT]):
    def __init__(self, state: ConnectionState, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._state = state

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    async def fetch(self, *, language: Language | None = None) -> FetchedBundle:
        """Fetch this bundle.

        Shorthand for:

        .. code:: python

            bundle = await client.fetch_bundle(bundle_id)
        """
        return await self._state.fetch_bundle(self.id, language)

    async def store_item(self, *, language: Language | None = None) -> BundleStoreItem:
        """Fetch the store item for this bundle.

        Shorthand for:

        .. code:: python

            (item,) = await client.fetch_store_item(bundles=[bundle], language=language)
        """
        from .store import BundleStoreItem

        language = language or self._state.http.language
        (item,) = await self._state.fetch_store_info(bundle_ids=(self.id,), language=language)
        return BundleStoreItem(self._state, item, language)

    async def apps(self) -> list[PartialApp[None]]:
        """Fetch the bundle's apps."""
        fetched = await self.fetch()
        return fetched._apps

    async def packages(self) -> list[PartialPackage[None]]:
        """Fetch the bundle's packages."""
        fetched = await self.fetch()
        return fetched._packages

    async def redeem(self) -> None:
        """Redeem this bundle if it's a free-on-demand bundle."""
        await self._state.http.redeem_bundle(self.id)


class FetchedBundle(PartialBundle[str]):
    """Represents a bundle of games and packages."""

    def __init__(self, state: ConnectionState, data: bundle.Bundle) -> None:
        super().__init__(state, id=data["bundleid"], name=data["name"])
        self._apps = [PartialApp(state, id=app_id) for app_id in data["appids"]]
        self._packages = [PartialPackage(state, id=package_id) for package_id in data["packageids"]]

        self._on_windows = data["available_windows"]
        self._on_mac_os = data["available_mac"]
        self._on_linux = data["available_linux"]
        self._vrhmod = data.get("support_vrhmod", False)
        self._vrhmod_only = data.get("support_vrhmod_only", False)
        self.creator_clans = [PartialClan(state, id) for id in data["creator_clan_ids"]]
        """The clans of the creators that created this bundle."""
        self.supported_languages = [Language.try_value(lang) for lang in data["localized_langs"]]
        """The languages that this bundle is supported in."""

    def is_on_windows(self) -> bool:
        """Whether the app is playable on Windows."""
        return self._on_windows

    def is_on_mac_os(self) -> bool:
        """Whether the app is playable on macOS."""
        return self._on_mac_os

    def is_on_linux(self) -> bool:
        """Whether the app is playable on Linux."""
        return self._on_linux

    async def apps(self) -> list[PartialApp[None]]:
        return self._apps

    async def packages(self) -> list[PartialPackage[None]]:
        return self._packages



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ..utils import StructIO

if TYPE_CHECKING:
    from typing_extensions import Self


class StructMessage:
    __slots__ = ()

    def from_dict(self, dict: dict[str, Any]) -> Self:
        self.__init__(**dict)
        return self

    def to_dict(self, *_: Any) -> dict[str, Any]:
        try:
            annotations = self.__annotations__
        except AttributeError:
            annotations = ()
        return {key: getattr(self, key) for key in annotations}

    def __bytes__(self) -> bytes:
        try:
            annotations = self.__annotations__
        except AttributeError:
            annotations = {}
        with StructIO() as io:
            for key, annotation in annotations.items():
                if isinstance(annotation, type):
                    raise RuntimeError("annotations shouldn't be types")
                if annotation in {"int", "AssetID"}:
                    io.write_u64(getattr(self, key))
                elif annotation == "bool":
                    io.write_u8(getattr(self, key))
                else:
                    raise RuntimeError(f"Unknown field {annotation}")

            return io.buffer

    def parse(self, data: bytes) -> Self:
        raise NotImplementedError
